"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

from .image_display import ImagePlotWidget



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for Scope controller

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import sys
import unittest
import configparser
import pvaccess as pva

from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import Parameter

from c2dataviewer.model import DataSource as DataReceiver
from c2dataviewer.control.scopeconfig import Configure
from c2dataviewer.scope import ScopeWindow, WarningDialog
from c2dataviewer.control import ScopeController


os.environ["QT_QPA_PLATFORM"] = "offscreen"


class TestImageDisplay(unittest.TestCase):
    """
    Units tests for the scope controllers.
    """

    @classmethod
    def setUpClass(cls):
        """
        Set up QApplication for all tests in this class.
        """
        cls.app = QtWidgets.QApplication.instance()
        if cls.app is None:
            cls.app = QtWidgets.QApplication(sys.argv)

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """

        # Create ImageWindow and get the imageWidget instance
        self.window = ScopeWindow()

        # Build parameter tree
        configure = Configure(configparser.ConfigParser())
        self.parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(self.parameters, showTop=False)

        # Model to be used
        self.model = DataReceiver()

        # Build GUI elements
        self.warning = WarningDialog(None)

        self.scope_controller = ScopeController(
            widget=self.window, model=self.model, parameters=self.parameters, WARNING=self.warning)

    def tearDown(self):
        """
        Tear down the environment after each test case.
        This method is called after each test.

        :return:
        """
        # Clean up the window to prevent Qt object deletion issues
        if hasattr(self, 'window'):
            self.window.close()
            self.window.deleteLater()
        if hasattr(self, 'scope_controller'):
            del self.scope_controller
        # Process pending Qt events to ensure cleanup
        QtWidgets.QApplication.processEvents()

    def create_controller(self, configure):
        """
        Helper method to create a controller with the given Configure object.

        :param configure: Configure object with parsed configuration
        :return: ScopeController instance
        """
        parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(parameters, showTop=False)

        controller = ScopeController(
            widget=self.window, model=self.model, parameters=parameters, WARNING=self.warning)

        return controller

    def test_update_status(self):
        """
        Test update statistics
        """
        #make sure that update_status runs without errors
        self.scope_controller.update_status()


    def test_get_fdr(self):
        """
        Test if the PV structure is properly parsed and correct fields are
        extracted to be used as a drop down items on the GUI. This are pva.ScalarType
        and arrays of pva.ScalarType.
        """
        # Mock get for the receiver
        def mock_get(*_):

            test_struct = {
                "anInt": pva.INT,
                "aStruct": {
                    "aStruct_aFloat": pva.FLOAT,
                    "aStruct_aDoubleArray": [pva.DOUBLE],
                },
                "aFloatArray": [pva.FLOAT],
                "aStringArray": [pva.STRING],
                "aStructArray": [
                    {
                        "aStructArray_aFloat": pva.FLOAT,
                        "aStructArray_aDoubleArray": [pva.DOUBLE],
                    }
                ],
                "aString": pva.STRING,
                "aDoubleArray": [pva.DOUBLE],
                "aVariant": (),
                "aBooleanArray": [pva.BOOLEAN],
            }
            return pva.PvObject(test_struct)
        self.scope_controller.model.get = mock_get

        fdr, fdr_scalar, fdr_nonnumeric = self.scope_controller.get_fdr()

        self.assertListEqual(fdr, ['aBooleanArray', 'aDoubleArray',
                             'aFloatArray', 'aStruct.aStruct_aDoubleArray'])
        self.assertListEqual(
            fdr_scalar, ['aString', 'aStruct.aStruct_aFloat', 'anInt'])

        self.assertListEqual(fdr_nonnumeric, ['aStringArray'])
        
    def test_buffer_size(self):
        #pass in data
        self.scope_controller.monitor_callback({ 'x' : [10]*100})
        bufval = self.scope_controller.parameters.child("Acquisition").child("Buffer (Samples)").value()
        self.assertEqual(bufval, 100)

        #set to object unit
        self.scope_controller.set_buffer_unit('Objects')
        bufval = self.scope_controller.parameters.child("Acquisition").child("Buffer (Objects)").value()
        self.assertEqual(bufval, 1)

    def test_config_file_settings(self):
        """
        Test that settings from config file are properly loaded into controller.
        """
        config_str = """
        [SCOPE]
        DefaultProtocol = ca
        AUTOSCALE = True

        [DISPLAY]
        MODE = fft
        AVERAGE = 5
        REFRESH = 200

        [ACQUISITION]
        BUFFER = 500
        """
        parser = configparser.ConfigParser()
        parser.read_string(config_str)

        configure = Configure(parser)
        controller = self.create_controller(configure)

        # Verify config settings are loaded
        self.assertTrue(controller.parameters.child("Display").child("Autoscale").value())
        self.assertEqual(controller.parameters.child("Display").child("Mode").value(), 'fft')
        self.assertEqual(controller.parameters.child("Display").child("Exp moving avg").value(), 5)
        self.assertEqual(controller.parameters.child("Display").child("Refresh").value(), 0.2)
        self.assertEqual(controller.parameters.child("Acquisition").child("Buffer (Samples)").value(), 500)

    def test_kwarg_settings(self):
        """
        Test that keyword arguments are properly used in controller setup.
        """
        parser = configparser.ConfigParser()
        parser.read_string("[SCOPE]\nDefaultProtocol = ca")

        configure = Configure(parser, arrayid='TestArray', xaxes='TestXAxis')
        controller = self.create_controller(configure)

        # Verify kwarg settings are used
        self.assertEqual(controller.parameters.child("Config").child("ArrayId").value(), 'TestArray')
        self.assertEqual(controller.parameters.child("Config").child("X Axes").value(), 'TestXAxis')

    def test_kwarg_overrides_config(self):
        """
        Test that keyword arguments override config file settings.
        """
        config_str = """
        [SCOPE]
        DefaultProtocol = ca

        [CONFIG]
        ARRAYID = ConfigArrayId
        XAXES = ConfigXAxis
        """
        parser = configparser.ConfigParser()
        parser.read_string(config_str)

        configure = Configure(parser, arrayid='KwargArrayId', xaxes='KwargXAxis')
        controller = self.create_controller(configure)

        # Verify kwargs override config file values
        self.assertEqual(controller.parameters.child("Config").child("ArrayId").value(), 'KwargArrayId')
        self.assertEqual(controller.parameters.child("Config").child("X Axes").value(), 'KwargXAxis')

    def test_pv_kwarg(self):
        """
        Test that PV can be set via keyword argument.
        """
        parser = configparser.ConfigParser()
        parser.read_string("[SCOPE]\nDefaultProtocol = ca")

        configure = Configure(parser, pv={'test': 'TEST:PV:NAME'})
        controller = self.create_controller(configure)

        # Verify PV kwarg is used
        self.assertEqual(controller.parameters.child("Acquisition").child("PV").value(), 'TEST:PV:NAME')

    def test_serialize_full_config(self):
        """
        Test that serialize function writes proper configuration to file with channels.
        """
        from io import StringIO
        from configparser import ConfigParser

        config_str = """
        [SCOPE]
        DefaultProtocol = pva

        [DISPLAY]
        MODE = fft
        AVERAGE = 10
        AUTOSCALE = True
        REFRESH = 500

        [ACQUISITION]
        PV = TEST:SCOPE:PV
        BUFFER = 1000
        BUFFERUNIT = Samples

        [CONFIG]
        ARRAYID = myArrayId
        XAXES = myXAxis
        MAJORTICKS = 5
        MINORTICKS = 2
        """
        parser = ConfigParser()
        parser.read_string(config_str)

        configure = Configure(parser)
        controller = self.create_controller(configure)

        # Set limits for Field parameters so they can accept custom values
        field_limits = ["None", "field1", "field2"]
        controller.parameters.child("Channel 1", "Field").setLimits(field_limits)
        controller.parameters.child("Channel 2", "Field").setLimits(field_limits)

        # Set some channel fields in parameters (serialize reads from parameters)
        controller.parameters.child("Channel 1", "Field").setValue("field1")
        controller.parameters.child("Channel 1", "DC offset").setValue(5.0)
        controller.parameters.child("Channel 2", "Field").setValue("field2")

        # Serialize to StringIO
        output = StringIO()
        controller.serialize(output)

        # Read back the serialized config
        output.seek(0)
        result_cfg = ConfigParser()
        result_cfg.read_file(output)

        # Verify app type
        self.assertEqual(result_cfg.get('DEFAULT', 'APP'), 'SCOPE')

        # Verify Display settings
        self.assertEqual(result_cfg.get('DISPLAY', 'mode'), 'fft')
        self.assertEqual(result_cfg.get('DISPLAY', 'average'), '10')
        self.assertEqual(result_cfg.get('DISPLAY', 'autoscale'), 'True')
        self.assertEqual(result_cfg.get('DISPLAY', 'refresh'), '500')

        # Verify Acquisition settings
        self.assertEqual(result_cfg.get('ACQUISITION', 'pv'), 'TEST:SCOPE:PV')
        self.assertEqual(result_cfg.get('ACQUISITION', 'buffer'), '1000')
        self.assertEqual(result_cfg.get('ACQUISITION', 'bufferunit'), 'Samples')

        # Verify Config settings
        self.assertTrue(result_cfg.has_section('CONFIG'))
        self.assertEqual(result_cfg.get('CONFIG', 'majorticks'), '5')
        self.assertEqual(result_cfg.get('CONFIG', 'minorticks'), '2')

        # Verify channel configurations
        self.assertTrue(result_cfg.has_section('CHANNELS'))
        self.assertEqual(result_cfg.get('CHANNELS', 'chan1.field'), 'field1')
        self.assertEqual(result_cfg.get('CHANNELS', 'chan1.dcoffset'), '5.0')
        self.assertEqual(result_cfg.get('CHANNELS', 'chan2.field'), 'field2')

    def test_serialize_minimal_config(self):
        """
        Test serialize with minimal configuration (default settings).
        """
        from io import StringIO
        from configparser import ConfigParser

        parser = ConfigParser()
        parser.read_string("[SCOPE]\nDefaultProtocol = ca")

        configure = Configure(parser)
        controller = self.create_controller(configure)

        # Serialize to StringIO
        output = StringIO()
        controller.serialize(output)

        # Read back the serialized config
        output.seek(0)
        result_cfg = ConfigParser()
        result_cfg.read_file(output)

        # Verify app type
        self.assertEqual(result_cfg.get('DEFAULT', 'APP'), 'SCOPE')

        # Verify Display section exists (should have default values)
        self.assertTrue(result_cfg.has_section('DISPLAY'))

        # Verify Acquisition section exists
        self.assertTrue(result_cfg.has_section('ACQUISITION'))

    def test_serialize_buffer_objects(self):
        """
        Test serialize with buffer unit set to Objects.
        """
        from io import StringIO
        from configparser import ConfigParser

        config_str = """
        [SCOPE]
        DefaultProtocol = pva

        [ACQUISITION]
        BUFFERUNIT = Objects
        BUFFER = 50
        """
        parser = ConfigParser()
        parser.read_string(config_str)

        configure = Configure(parser)
        controller = self.create_controller(configure)

        # Serialize to StringIO
        output = StringIO()
        controller.serialize(output)

        # Read back the serialized config
        output.seek(0)
        result_cfg = ConfigParser()
        result_cfg.read_file(output)

        # Verify buffer unit and size
        self.assertEqual(result_cfg.get('ACQUISITION', 'bufferunit'), 'Objects')
        self.assertEqual(result_cfg.get('ACQUISITION', 'buffer'), '50')




# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2020 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Helper functions for c2d unit tests.

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
from datetime import datetime as dt
import numpy as np
import pvaccess as pva
from pyqtgraph.Qt import QtWidgets

from c2dataviewer.view.image_definitions import *
from c2dataviewer.view.image_display import ImagePlotWidget

############################################
# ND Image
############################################
def get_time_stamp(time_stamp=None):
    """
    Transform seconds to PVA timestamp (or generate for the current time if time_stamp=None)

    :param time_stamp: (int) Timestamp in nano seconds or None to take current time.
    :return: (PvTimeStamp) Timestramp value.
    """
    NANOSECONDS_IN_SECOND = 1000000000
    if time_stamp is None:
        time_stamp = np.datetime64(dt.now(), 'ns')
    else:
        time_stamp = np.datetime64(time_stamp, 'ns')
    t = (time_stamp-np.datetime64(0,'ns'))/np.timedelta64(1, 's')
    s = int(t)
    ns = int((t - s)*NANOSECONDS_IN_SECOND)
    return pva.PvTimeStamp(s,ns,0)

def create_image(id, image=None, data_type='ubyteValue', nx=None, ny=None, nz=None, color_mode=None, time_stamp=None, extra_fields_PV_object=None):
    """
    Generate image as NtNdArray.

    :param id: (int) Array index.
    :param index: (np.array) Image.
    :param data_type: (string) Datatype of the image. Default value is 'ubyteValue'.
    :param nx: (int) Dimension of the image in X direction.
    :param ny: (int) Dimension of the image in Y direction.
    :param ny: (int) Dimension of the image in Z direction. Should be None (default) or 3.
    :param color_mode: (int) Color mode the image. Possible values are 0, 2, 3, 4.
    :param time_stamp: (int) EPICS timestamp in seconds. Can be None in which case the current time is used.
    :param extra_fields_PV_object: (dict) Extra fields which will be added to the NtNdArray. Keys must strings.
    """
    # Generate timestamp and the image
    time_stamp = get_time_stamp(time_stamp)
    if image is None:
        image = np.random.randint(0,256, size=nx*ny, dtype=np.uint8)

    # Build the NtNdArray
    if extra_fields_PV_object:
        nda = pva.NtNdArray(extra_fields_PV_object)
    else:
        nda = pva.NtNdArray()
    nda['uniqueId'] = id
    dims = [pva.PvDimension(nx, 0, nx, 1, False), pva.PvDimension(ny, 0, ny, 1, False)]
    if nz is not None:
        if color_mode == COLOR_MODE_RGB1:
            dims.insert(0, pva.PvDimension(nz, 0, nz, 1, False))
        elif color_mode == COLOR_MODE_RGB2:
            dims.insert(1, pva.PvDimension(nz, 0, nz, 1, False))
        elif color_mode == COLOR_MODE_RGB3:
            dims.append(pva.PvDimension(nz, 0, nz, 1, False))
    nda['dimension'] = dims
    nda['descriptor'] = 'PvaPy Simulated Image'
    nda['compressedSize'] = nx*ny*(nz if nz is not None else 1)
    nda['uncompressedSize'] = nx*ny*(nz if nz is not None else 1)
    nda['timeStamp'] = time_stamp
    nda['dataTimeStamp'] = time_stamp
    if color_mode is not None:
        attrs = [pva.NtAttribute('ColorMode', pva.PvInt(color_mode))]
        nda['attribute'] = attrs
    nda['value'] = {data_type : image}

    return nda

app = None

def setup_qt_app():
    global app
    
    if not app:
        app = QtWidgets.QApplication.instance()
        if not app:
            app  = QtWidgets.QApplication([])



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2021 UChicago Argonne LLC
 as operator of Argonne National Laboratory

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import numpy as np

# Area detector color modes values. Source: https://github.com/areaDetector/ADCore/blob/master/ADApp/ADSrc/NDArray.h#L29
# Currently only four are supported.
COLOR_MODE_MONO = 0 # [NX, NY]
COLOR_MODE_RGB1 = 2 # [3, NX, NY]
COLOR_MODE_RGB2 = 3 # [NX, 3, NY]
COLOR_MODE_RGB3 = 4 # [NX, NY, 3]

COLOR_MODES = {
    COLOR_MODE_MONO : "MONO",
    COLOR_MODE_RGB1 : "RGB1",
    COLOR_MODE_RGB2 : "RGB2",
    COLOR_MODE_RGB3 : "RGB3",
}

def transcode_image(image, color_mode, x, y, z):
    """
    Transcode the numpy array to the correct shape.
    For MONO images this is [NX, NY] and for the RGB [NX, NY, 3] as documented:
    https://github.com/pyqtgraph/pyqtgraph/blob/ea08dda62dade523f2193ec353b9bacac6d8d35d/pyqtgraph/widgets/RawImageWidget.py#L41

    :param image: (numpy array) Flat image to be trancoded.
    :param color_mode: (enum) Color mode of the image.
    :param x: (int) X dimension of the image.
    :param y: (int) Y dimension of the image.
    :param z: (int) Z dimension of the image. Can be None for mono images.
    :return: (numpy array) Transcoded np.array in either [NX, NY] or [NX, NY, 3] shapes.
    """

    if color_mode == COLOR_MODE_MONO:
        return np.reshape(image, (y, x))

    elif color_mode == COLOR_MODE_RGB1:
        image = np.reshape(image, (y, x, z))
        return image

    elif color_mode == COLOR_MODE_RGB2:
        image = np.reshape(image, (y, z, x))
        image = np.swapaxes(image, 2, 1)
        return image

    elif color_mode == COLOR_MODE_RGB3:
        image = np.reshape(image, (z, y, x))
        image = np.swapaxes(image, 0, 2)
        image = np.swapaxes(image, 0, 1)
        return image

    else:
        raise RuntimeError("Unsupported color mode.")



from collections import namedtuple
from enum import Enum, auto
from configparser import ConfigParser
import re
import logging


class AppType(Enum):
    SCOPE = "scope"
    IMAGE = "image"
    STRIPTOOL = "striptool"

    def __str__(self):
        return self.value

class Scope(Enum):
    BUFFER = auto()
    BUFFER_UNIT = auto()
    DISPLAY_MODE = auto()
    FFT_FILTER = auto()
    AVERAGE = auto()
    AUTOSCALE = auto()
    SINGLE_AXIS = auto()
    HISTOGRAM = auto()
    NBIN = auto()
    REFRESH = auto()
    TRIGGER = auto()
    TRIGGER_MODE = auto()
    TRIGGER_THRESHOLD = auto()
    TRIGGER_AUTOSCALE_BUFFER = auto()
    TRIGGER_TIME_FIELD = auto()
    TRIGGER_DATA_TIME_FIELD = auto()
    CHANNEL_COUNT = auto()
    MOUSE_OVER = auto()
    CONNECT_ON_START = auto()
    PV = auto()
    ARRAYID = auto()
    XAXES = auto()
    MAJORTICKS = auto()
    MINORTICKS = auto()
    EXTRA_DISPLAY_FIELDS = auto()
    MOUSE_OVER_DISPLAY_LOCATION = auto()

class Striptool(Enum):
    DEFAULT_PROTOCOL = auto()
    SAMPLEMODE = auto()
    
Field = namedtuple('Field', ['loc', 'type'])
MultiLoc = namedtuple('MultiLoc', ['sections', 'write_section', 'field'])

_schema = {
    Scope.BUFFER : Field(loc=('ACQUISITION', 'BUFFER'),type=int),
    Scope.BUFFER_UNIT: Field(loc=('ACQUISITION', 'BUFFERUNIT'),type=['samples', 'objects']),
    Scope.AUTOSCALE: Field(loc=MultiLoc(['STRIPTOOL', 'SCOPE', 'DISPLAY'], 'DISPLAY', 'AUTOSCALE'), type=bool),
    Scope.DISPLAY_MODE: Field(loc=('DISPLAY', 'MODE'), type=['normal', 'fft', 'psd', 'diff', 'autocorrelate_fft']),
    Scope.FFT_FILTER: Field(loc=('DISPLAY', 'FFT_FILTER'), type=['none', 'hamming']),
    Scope.AVERAGE: Field(loc=('DISPLAY', 'AVERAGE'), type=int),
    Scope.SINGLE_AXIS: Field(loc=('DISPLAY', 'SINGLE_AXIS'), type=bool),
    Scope.HISTOGRAM: Field(loc=('DISPLAY', 'HISTOGRAM'), type=bool),
    Scope.NBIN: Field(loc=('DISPLAY', 'N_BIN'), type=int),
    Scope.REFRESH: Field(loc=('DISPLAY', 'REFRESH'), type=int),
    Scope.TRIGGER: Field(loc=('TRIGGER', 'TRIGGER'), type=str),
    Scope.TRIGGER_MODE: Field(loc=('TRIGGER', 'TRIGGER_MODE'),
                              type=['none', 'onchange', 'gtthreshold', 'ltthreshold']),
    Scope.TRIGGER_THRESHOLD: Field(loc=('TRIGGER', 'THRESHOLD'), type=float),
    Scope.TRIGGER_AUTOSCALE_BUFFER: Field(loc=('TRIGGER', 'AUTOSCALE_BUFFER'), type=bool),
    Scope.TRIGGER_TIME_FIELD: Field(loc=('TRIGGER', 'TIME_FIELD'), type=str),
    Scope.TRIGGER_DATA_TIME_FIELD: Field(loc=('TRIGGER', 'DATA_TIME_FIELD'), type=str),
    Scope.CHANNEL_COUNT: Field(loc=('CHANNELS', 'COUNT'), type=int),
    Scope.MOUSE_OVER: Field(loc=('DISPLAY', 'MOUSEOVER'), type=bool),
    Scope.CONNECT_ON_START: Field(loc=('ACQUISITION', 'CONNECTONSTART'), type=bool),
    Scope.PV: Field(loc=('ACQUISITION', 'PV'), type=str),
    Scope.ARRAYID: Field(loc=('CONFIG', 'ARRAYID'), type=str),
    Scope.XAXES: Field(loc=('CONFIG', 'XAXES'), type=str),
    Scope.MAJORTICKS: Field(loc=('CONFIG', 'MAJORTICKS'), type=int),
    Scope.MINORTICKS: Field(loc=('CONFIG', 'MINORTICKS'), type=int),
    Scope.EXTRA_DISPLAY_FIELDS: Field(loc=('CONFIG', 'EXTRADISPLAYFIELDS'), type=list),
    Scope.MOUSE_OVER_DISPLAY_LOCATION: Field(loc=('CONFIG', 'EXTRADISPLAYLOCATION'), type=['top_right', 'bottom_right', 'bottom_left']),
    Striptool.DEFAULT_PROTOCOL: Field(loc=('STRIPTOOL', 'DEFAULTPROTOCOL'), type=str),
    Striptool.SAMPLEMODE: Field(loc=('ACQUISITION', 'SAMPLEMODE'), type=str)
}

class Parser:
    def __init__(self, cfg):
        self.cfg = cfg
        self.logger = logging.getLogger()
        
    def get_sections_list(self):
        appname = self.get_appname()
        try:
            sections = self.cfg[appname]['SECTIONS']
        except:
            raise ValueError(f'"SECTIONS" setting not found in {appname}')
        return sections.split()

    def get_channel_config(self, is_striptool=False):
        section = None
        try:
            if is_striptool:
                section=self.cfg['STRIPTOOL']
            else:
                section=self.cfg['CHANNELS']               
        except:
            section = None
            
        #Read channel information.  Channel order is
        #determined by order in config
        chan_cfg_lookup = {}
        
        if section:
            for k, v in section.items():
                if bool(re.match('chan', k, re.I)):
                    ch, param = k.lower().split('.')
                    chan_cfg_lookup[ch] = chan_cfg_lookup.get(ch, {})
                    chan_cfg_lookup[ch][param] = v

        chan_cfgs = list(chan_cfg_lookup.values())
        return chan_cfgs
    
    def get_appname(self):
        return self.cfg.get('DEFAULT', 'APP')        
        
    def get(self, key, default=None):
        try:
            fielddef = _schema[key]
        except Exception as e:
            raise KeyError('Invalid configuration key') from e            

        loc=fielddef.loc
        
        if isinstance(loc, MultiLoc):
            for s in loc.sections:
                l = (s, loc.field)
                val = self.__get(l, fielddef, None)
                if val is not None:
                    return val
            return default
        else:
            val =  self.__get(loc, fielddef, default)
            return val
        
    def __get(self, loc, fielddef, default):
        try:
            if fielddef.type == int:
                val = self.cfg.getint(loc[0], loc[1], fallback=default)
            elif fielddef.type == float:
                val = self.cfg.getfloat(loc[0], loc[1], fallback=default)
            elif fielddef.type == bool:
                val = self.cfg.getboolean(loc[0], loc[1], fallback=default)
            elif fielddef.type == str:
                val = self.cfg.get(loc[0], loc[1], fallback=default)
            elif fielddef.type == list:
                val = self.cfg.get(loc[0], loc[1], fallback=default)
                val = val.lower().strip().split(',') if val else []            
            elif isinstance(fielddef.type, list):
                val = self.cfg.get(loc[0], loc[1], fallback=default)
                val = val.lower().strip() if val else val
                if val is not None and val not in fielddef.type:                
                    raise ValueError(f'Invalid for key {loc[1]}: value {val}')
            else:
                raise ValueError(f'Invalid type: {fielddef.type}')
        except Exception as e:
            self.logger.error(f'Failed to get {loc}:{e}')
            val = default
        return val
    
class Serializer:
    def __init__(self):
        self.cfg = ConfigParser()

    def set(self, key, value):
        try:
            fielddef = _schema[key]
        except Exception as e:
            raise KeyError('Invalid configuration key') from e

        loc=fielddef.loc
        if isinstance(loc, MultiLoc):
            section = loc.write_section
            option = loc.field
        else:
            section, option = loc[0], loc[1]

        # Ensure section exists
        if not self.cfg.has_section(section):
            self.cfg.add_section(section)

        self.cfg.set(section, option, str(value))

    def set_app(self, app: AppType):
        self.__set_raw('DEFAULT', 'APP', str(app).upper())

    def __set_raw(self, section, option, value):
        '''
        Set a configuration value directly without schema validation.
        Useful for app-specific or channel-specific settings.
        '''
        # DEFAULT section is special in ConfigParser and cannot be added
        if section != 'DEFAULT' and not self.cfg.has_section(section):
            self.cfg.add_section(section)
        self.cfg.set(section, option, str(value))

    def write_channels(self, section: str, chan_cfgs: list[dict]):
        '''
        Write channel configurations to a specific section.

        section: Section name to write channels to (e.g., 'STRIPTOOL', 'CHANNELS')
        chan_cfgs: List of channel configs, where a channel config
                   is a dictionary of channel fields
        '''
        if not self.cfg.has_section(section):
            self.cfg.add_section(section)

        i = 1
        for cfg in chan_cfgs:
            prefix = f'chan{i}.'
            for k, v in cfg.items():
                key = f'{prefix}{k.lower()}'
                self.cfg.set(section, key, str(v))
            i += 1

    def write(self, cfile):
        self.cfg.write(cfile)



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import sys
import os
import psutil
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import uic
from pyqtgraph.parametertree import Parameter
from .model import DataSource as DataReceiver
from .control import StripToolController

from .control.striptoolconfig import StripToolConfigure

form_path = os.path.join(os.path.dirname(__file__), "ui/striptool.ui")
form_class = uic.loadUiType(form_path)[0]

class StripToolWindow(QtWidgets.QMainWindow, form_class):
    def __init__(self, parent=None):
        super(StripToolWindow, self).__init__(parent=parent)
        self._proc = psutil.Process()
        self.setupUi(self)
        self.show()

pvedit_dialog_path = os.path.join(os.path.dirname(__file__), "ui/addpv.ui")
pvedit_dialog_class = uic.loadUiType(pvedit_dialog_path)[0]

class PvEditDialog(QtWidgets.QDialog, pvedit_dialog_class):
    def __init__(self, parent=None):
        super(PvEditDialog, self).__init__(parent=parent)
        self.setupUi(self)

warning_path = os.path.join(os.path.dirname(__file__), "ui/warning.ui")
warning_class = uic.loadUiType(warning_path)[0]

class WarningDialog(QtWidgets.QDialog, warning_class):
    def __init__(self, parent=None):
        super(WarningDialog, self).__init__(parent=parent)
        self.setupUi(self)

def striptool(cfg, **kwargs):
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    else:
        print('QApplication instance already exists: %s' % str(app))

    w = StripToolWindow()
    model = DataReceiver()
    warning = WarningDialog()
    pvedit_dialog = PvEditDialog()

    configure = StripToolConfigure(cfg, **kwargs)
    parameters = Parameter.create(
        name="params", type="group", children=configure.parse())
    w.parameterPane.setParameters(parameters, showTop=False)

    controller = StripToolController(w, model, pvedit_dialog, warning, parameters, configure, **kwargs)
    parameters.sigTreeStateChanged.connect(controller.parameter_change)
    
    w.show()
    try:
        app.exec_()
    finally:
        controller.stop_plotting()



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""

import sys
import os.path
import psutil
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import uic
from pyqtgraph.parametertree import Parameter

# Updated import: Configure now lives in control.scopeconfig
from .control.scopeconfig import Configure
from .model import DataSource as DataReceiver
from .control import ScopeController

form_path = os.path.join(os.path.dirname(__file__), "ui/scope.ui")
form_class = uic.loadUiType(form_path)[0]


class ScopeWindow(QtWidgets.QMainWindow, form_class):
    def __init__(self, parent=None):
        super(ScopeWindow, self).__init__(parent=parent)
        self._proc = psutil.Process()
        self.setupUi(self)
        self.show()


warning_path = os.path.join(os.path.dirname(__file__), "ui/warning.ui")
warning_class = uic.loadUiType(warning_path)[0]


class WarningDialog(QtWidgets.QDialog, warning_class):
    def __init__(self, parent=None):
        super(WarningDialog, self).__init__(parent=parent)
        self.setupUi(self)


def scope(cfg, **kwargs):
    """
    Main function for scope display

    :return:
    """

    # Check for an instance of a QtWidgets.QApplication, if so use it...
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    else:
        print('QApplication instance already exists: %s' % str(app))

    w = ScopeWindow()

    configure = Configure(cfg, **kwargs)
    parameters = Parameter.create(
        name="params", type="group", children=configure.parse())
    w.parameterPane.setParameters(parameters, showTop=False)
    pvmap = configure.pvs

    warning = WarningDialog(None)

    default_pv = list(pvmap.values())[0] if pvmap else None    
    model = DataReceiver(default=default_pv)
    controller = ScopeController(w, model, parameters, WARNING=warning)
    controller.default_config(**kwargs)

    parameters.sigTreeStateChanged.connect(controller.parameter_change)

    w.show()
    try:
        app.exec_()
    finally:
        controller.stop_plotting()



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities for image display

@author: Guobao Shen <gshen@anl.gov>
"""
from collections import namedtuple
import queue
import logging
import math
import numpy as np
import pyqtgraph as pg
from pyqtgraph import QtCore
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.widgets.RawImageWidget import RawImageWidget
# We attempt to import modules needed for compressed images
# and ignore errors if they are not there
try:
    import blosc
except ImportError:
    pass
try:
    import lz4.block
except ImportError:
    pass
try:
    import bitshuffle
except ImportError:
    pass

import pvaccess as pva

from .image_definitions import *
from .image_profile_display import ImageProfileWidget
from .ui_components import TransparentRubberBand, RoiMidLines

class ImageCompressionUtility:

    # Compressed payload parameters for images compressed using AD codecs
    LZ4_PAYLOAD_START = 0
    BSLZ4_PAYLOAD_START = 0
    BSLZ4_BLOCK_SIZE = 0

    NUMPY_DATA_TYPE_MAP = {
        pva.UBYTE   : np.dtype('uint8'),
        pva.BYTE    : np.dtype('int8'),
        pva.USHORT  : np.dtype('uint16'),
        pva.SHORT   : np.dtype('int16'),
        pva.UINT    : np.dtype('uint32'),
        pva.INT     : np.dtype('int32'),
        pva.ULONG   : np.dtype('uint64'),
        pva.LONG    : np.dtype('int64'),
        pva.FLOAT   : np.dtype('float32'),
        pva.DOUBLE  : np.dtype('float64')
    }
    
    NTNDA_DATA_TYPE_MAP = {
        pva.UBYTE   : 'ubyteValue',
        pva.BYTE    : 'byteValue',
        pva.USHORT  : 'ushortValue',
        pva.SHORT   : 'shortValue',
        pva.UINT    : 'uintValue',
        pva.INT     : 'intValue',
        pva.ULONG   : 'ulongValue',
        pva.LONG    : 'longValue',
        pva.FLOAT   : 'floatValue',
        pva.DOUBLE  : 'doubleValue',
    }

    @classmethod
    def get_ntnda_data_type(cls, pvaType):
       return cls.NTNDA_DATA_TYPE_MAP.get(pvaType)

    @classmethod
    def get_decompressor(cls, codecName):
        utilityMap = {
            'blosc' : cls.blosc_decompress,
            'bslz4' : cls.bslz4_decompress,
            'lz4' : cls.lz4_decompress
        }
        decompressor = utilityMap.get(codecName)
        if not decompressor:
            raise RuntimeError(f'Unsupported compression: {codecName}')
        return decompressor

    @classmethod
    def blosc_decompress(cls, inputArray, inputType, uncompressedSize):
        try:
            oadt = cls.NUMPY_DATA_TYPE_MAP.get(inputType)
            oasz = uncompressedSize // oadt.itemsize
            outputArray = np.empty(oasz, dtype=oadt)
            nBytesWritten = blosc.decompress_ptr(bytearray(inputArray), outputArray.__array_interface__['data'][0])
        except NameError as ex:
            print(f'Required module for blosc_decompress is missing: {ex}')
            raise
        except Exception as ex:
            print(f'Error in blosc_decompress: {ex}')
            raise
        return outputArray

    @classmethod
    def lz4_decompress(cls, inputArray, inputType, uncompressedSize):
        try:
            oadt = cls.NUMPY_DATA_TYPE_MAP.get(inputType)
            outputBytes = lz4.block.decompress(inputArray.tobytes()[cls.LZ4_PAYLOAD_START:], uncompressed_size=uncompressedSize)
            outputArray = np.frombuffer(outputBytes, dtype=oadt)
        except NameError as ex:
            print(f'Required module for lz4_decompress is missing: {ex}')
            raise
        except Exception as ex:
            print(f'Error in lz4_decompress: {ex} {type(ex)}')
            raise
        return outputArray

    @classmethod
    def bslz4_decompress(cls, inputArray, inputType, uncompressedSize):
        try:
            oadt = cls.NUMPY_DATA_TYPE_MAP.get(inputType)
            oasz = uncompressedSize // oadt.itemsize
            oash = (oasz,)
            outputArray = bitshuffle.decompress_lz4(inputArray[cls.BSLZ4_PAYLOAD_START:], oash, oadt, cls.BSLZ4_BLOCK_SIZE)
        except NameError as ex:
            print(f'Required module for bslz4_decompress is missing: {ex}')
            raise
        except Exception as ex:
            print(f'Error in bslz4_decompress: {ex} {type(ex)}')
            raise
        return outputArray

class MouseDialog:
    def __init__(self, imageWidget):
        # mouse dialog flags        
        self.mouse_dialog_enabled = False
        self.mouse_dialog_launched = False
        self.mouse_dialog_hide = False
        
        self.max_textbox_width = 0

        # pixel coordinates
        self.pix_x = 0
        self.pix_y = 0

        # textbox coordinates inside the widget screen
        self.x_coor = 0
        self.y_coor = 0

        # intensity coordinates
        self.int_x = 0
        self.int_y = 0

        # initialize mouse dialog widgets
        self.layout = QtWidgets.QVBoxLayout(imageWidget)
        self.textbox = QtWidgets.QTextEdit(imageWidget)
        self.textbox.setCursorWidth(0)
        self.textbox.setFixedSize(0,0)

    def enable_mouse_dialog(self):
        self.mouse_dialog_enabled = True
        self.textbox.setHidden(False)

    def disable_mouse_dialog(self):
        self.mouse_dialog_enabled = False
        self.mouse_dialog_launched = False
        self.textbox.setHidden(True)

Image = namedtuple("Image", ['id', 'new', 'image', 'black', 'white'])
class ImagePlotWidget(RawImageWidget):

    MIN_DISPLAY_QUEUE_SIZE = 0
    MAX_DISPLAY_QUEUE_SIZE = 999
    DEFAULT_DISPLAY_QUEUE_SIZE = 20
    ZOOM_LENGTH_MIN = 4 # Using zoom this is the smallest number of pixels to display in each direction

    _set_image_signal = QtCore.pyqtSignal()
    connection_changed_signal = QtCore.Signal(str, str)
    right_button_clicked_signal = QtCore.pyqtSignal(QtCore.QPoint)
    
    def __init__(self, parent=None, **kargs):
        RawImageWidget.__init__(self, parent, scaled=True)

        self._set_image_signal.connect(self._set_image_signal_callback)

        self._noagc = kargs.get("noAGC", False)
        # self.camera_changed()
        self._agc = False
        self._lastTimestamp = None
        self._freeze = False

        self.__last_array_id = None
        self.dataType = None
        self.MB_received = 0.0
        self.frames_received = 0
        self.frames_displayed = 0
        self.frames_missed = 0

        # Image properties
        self.dimensions = None
        self.color_mode = None
        self.x = None
        self.y = None
        self.z = None

        # max value of image pixel
        self.maxVal = 0
        self.minVal = 0
        self.mutex = QtCore.QMutex()

        self._white = 255.0
        self._black = 0.0

        # Image profiles
        self.image_profile_widget = None

        # Moving average
        self.image_list = []
        self.moving_average_enabled = False
        self.n_moving_average_frames = 0

        # Last displayed image
        self.last_displayed_image = None

        # Widget properties
        self.image_width_pixels = 0
        self.image_height_pixels = 0

        # Zoom/ROI widgets
        self.roi_mode = False
        self.rulers_displayed = False
        self.roi_origin = None
        self.roi_mid_lines = None
        self.__zoomSelectionIndicator = None
        self.createZoomSelectionIndicator()
        self.__zoomDict = {
            "isZoom": False,
            'xoffset': 0,
            'yoffset': 0,
            'width': 0,
            'height': 0,
        }
        self.right_button_pressed = False
        self.is_image_panning = False

        # Limit control to avoid overflow network for best performance
        self._pref = {"Max": 0,
                      "Min": 0,
                      "DPX": 0,
                      "DPXEnabled": False,
                      "DPXLimit": 0x00,
                      "EmbeddedDataLen" : 0,
                      "CPU": 0,
                      "EnableCPULimit": False,
                      "CPULimit": 50,
                      "Net": 0,
                      "EnableNetLimit": False,
                      "NetLimit": 100,
                      "FPS": 0,
                      "FPSLimit": None,
                      "EnableServerQueue": False,
                      "ServerQueueSize": 10}

        self._isInputValid = False
        self._inputType = ""
        self.dead_pixels = [0]
        self._max = [0]
        self._min = [0]
        self.dead_pixels_roi = [0]
        self._maxRoi = [0]
        self._minRoi = [0]

        self._scaling = 1.0
        self._agc = True
        self._lastTimestamp = None

        # Imagecontroller callbacks
        self.updateGuiBlackLimits = None
        self.updateGuiWhiteLimits = None
        self.getGuiBlackLimits = None
        self.getGuiWhiteLimits = None
        self.updateGuiBlack = None
        self.updateGuiWhite = None

        self.datasource = None
        self.data = None

        self.dataTypesDict = {
            'byteValue' :   {'minVal' : int(-2**8 / 2),   'maxVal' : int(2**8 / 2 - 1),  'npdt' : "int8",   },
            'ubyteValue' :  {'minVal' : 0           ,     'maxVal' : int(2**8 - 1),      'npdt' : "uint8",  },
            'shortValue' :  {'minVal' : int(-2**16 / 2),  'maxVal' : int(2**16 / 2 - 1), 'npdt' : "int16",  },
            'ushortValue' : {'minVal' : 0               , 'maxVal' : int(2**16 - 1),     'npdt' : "uint16", },
            'intValue' :    {'minVal' : int(-2**32 / 2),  'maxVal' : int(2**32 / 2 - 1), 'npdt' : "int32",  },
            'uintValue' :   {'minVal' : 0               , 'maxVal' : int(2**32 - 1),     'npdt' : "uint32", },
            'longValue' :   {'minVal' : int(-2**64 / 2),  'maxVal' : int(2**64 / 2 - 1), 'npdt' : "int64",  },
            'ulongValue' :  {'minVal' : 0               , 'maxVal' : int(2**64 - 1),     'npdt' : "uint64", },
            'floatValue' :  {'minVal' : int(-2**24),      'maxVal' : int(2**24),         'npdt' : "float32",},
            'doubleValue' : {'minVal' : int(-2**53),      'maxVal' : int(2**53),         'npdt' : "float64",},
        }

        # Queue used to transfer the images from the process to the display thread
        self.draw_queue = queue.Queue(ImagePlotWidget.DEFAULT_DISPLAY_QUEUE_SIZE)

        # mouse dialog box
        self.mouse_dialog = MouseDialog(self)

    def rulersDisplayed(self, rulers_displayed):
        """
        Flag to keep track if rulers are displayed or not.

        :param rulers_displayed: (boolean) flag value
        :return:
        """
        self.rulers_displayed = rulers_displayed

    def createZoomSelectionIndicator(self, roi_mode=False):
        """
        Create zoom selection indicator based on the ROI mode.

        :param roi_mode: (boolean) Object holding information about the event.
        :return:
        """
        if self.__zoomSelectionIndicator:
            self.__zoomSelectionIndicator.hide()
        if self.roi_mid_lines:
            self.roi_mid_lines.hide()
            self.roi_mid_lines = None
        self.roi_mode = roi_mode
        if not roi_mode:
            self.__zoomSelectionIndicator = QtWidgets.QRubberBand(QtWidgets.QRubberBand.Rectangle, self)
        else:
            self.__zoomSelectionIndicator = TransparentRubberBand(QtWidgets.QRubberBand.Rectangle, self)

    def resizeEvent(self, event):
        """
        This method is called by the Qt when the widget change size. We use this to recalculate the
        size of the image od the display.

        :param event: (QResizeEvent) Object holding information about the event.
        :return:
        """
        self.calc_img_size_on_screen()

    def mousePressEvent(self, event):
        """
        This method is event handler for the mouse click event and is called by the Qt framework.

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        # Get location of the click
        click_position = event.pos()
        self.click_x = click_position.x()
        self.click_y = click_position.y()
        
        # Only triggered is left mouse button clicked 
        if event.button() == QtCore.Qt.LeftButton:
            # Return if there is no image data
            if not self.x or not self.y:
                return
            
            # Check if the press happened on the image
            img_width, img_height, _ = self.calc_img_size_on_screen()
            if (self.click_x > img_width or self.click_y > img_height):
                return

            # Mouse buttom was pressed on the image. We start panning.
            self.roi_origin = click_position
            self.__zoomSelectionIndicator.setGeometry(QtCore.QRect(self.roi_origin, QtCore.QSize()))
            self.__zoomSelectionIndicator.show()
            if self.roi_mode and self.roi_mid_lines:
                self.roi_mid_lines.hide()
                self.roi_mid_lines = None
        elif event.button() == QtCore.Qt.RightButton:
            # Flag to indicate click occured
            self.right_button_pressed = True

    def mouseMoveEvent(self, event):
        """
        This method is event handler for the mouse move event and is called by the Qt framework.

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        if self.mouse_dialog.mouse_dialog_enabled:
            self.updateMouseDialog(event)

        # panning zoomed image window
        if self.is_zoomed() and self.right_button_pressed:
            self.is_image_panning = True
            self.pan_zoom_window(event)

        # Mouse is moving, while selecting roi. Redraw the selection rectangle.
        if self.roi_origin is not None:
            self.__zoomSelectionIndicator.setGeometry(
                QtCore.QRect(self.roi_origin, event.pos()).normalized())

    def mouseReleaseEvent(self, event):
        """
        This method is event handler for the mouse button release event and is called by the Qt framework.

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        # Trigger context menu when not panning
        if event.button() == QtCore.Qt.RightButton:
            self.right_button_pressed = False
            if not self.is_image_panning:
                self.right_button_clicked_signal.emit(event.pos())
            self.is_image_panning = False

        if self.roi_origin is None:
            return

        # Get information about the widget dimensions
        panEnd = QtCore.QPoint(event.pos())
        imageGeometry = self.geometry().getRect()
        widget_width = imageGeometry[2]
        widget_hight = imageGeometry[3]

        # Calculate x parameters, x min/max and size in pixels
        xmin = self.roi_origin.x()
        xmax = panEnd.x()
        if xmin > xmax:
            xmax, xmin = xmin, xmax
        if xmin < 0:
            xmin = 0
        if xmax > widget_width:
            xmax = widget_width

        # Calculate y parameters, y min/max and size in pixels
        ymin = self.roi_origin.y()
        ymax = panEnd.y()
        if ymin > ymax:
            ymax, ymin = ymin, ymax
        if ymin < 0:
            ymin = 0
        if ymax > widget_hight:
            ymax = widget_hight

        self.roi_origin = None

        if self.roi_mode:
            if self.rulers_displayed:
                # Draw mid lines from ROI rectangle towards rulers
                xleft = xmin
                yleft = (ymin+ymax)/2.0
                xtop = (xmin+xmax)/2.0
                ytop = ymin
                self.roi_mid_lines = RoiMidLines(xleft, yleft, xtop, ytop, self)

                # Use image geometry for the ROI lines
                w = self.width()
                h = self.height()
                rectangle = QtCore.QRect(0, 0, w, h)
                self.roi_mid_lines.setGeometry(rectangle)
                self.roi_mid_lines.show()
            return

        # We made the roi selection
        # Hide selection rectangle
        self.__zoomSelectionIndicator.hide()

        # Calculate zoomed image parameters
        self.__calculateZoomParameters(xmin, xmax, ymin, ymax)
    
    def wheelEvent(self, event):
        """
        This method checks the angleDelta to determine the direction of the scroll to then
        set zoom region

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        # Calculate the position of the mouse cursor relative to the widget
        mouse_x = event.x()
        mouse_y = event.y()

        # Get current zoom parameters
        xOffset, yOffset, width, height = self.get_zoom_region()

        # Calculate the scale factor for zooming based on the mouse wheel delta
        scale_factor = 0.95 if event.angleDelta().y() > 0 else 1/0.95

        # Calculate the new width and height after zooming
        new_width = width * scale_factor
        new_height = height * scale_factor

        # Limit the minimum and maximum size after zooming
        min_size = 4
        
        # Zooming out/in
        if (scale_factor > 1):
            max_width_size = self.x
            max_height_size = self.y
            new_width = math.ceil(max(min(new_width, max_width_size), min_size))
            new_height = math.ceil(max(min(new_height, max_height_size), min_size))
        else:
            max_width_size = width
            max_height_size = height
            new_width = int(max(min(new_width, max_width_size), min_size))
            new_height = int(max(min(new_height, max_height_size), min_size))

        # Calculate the relative position of the mouse cursor within the widget
        if not width or not height:
            return
        relative_mouse_x = mouse_x / width
        relative_mouse_y = mouse_y / height

        # Calculate the new position of the mouse cursor after zooming
        new_mouse_x = relative_mouse_x * new_width
        new_mouse_y = relative_mouse_y * new_height

        # Calculate the new position of the widget after zooming to keep the mouse cursor centered
        new_x = int(xOffset + mouse_x - new_mouse_x)
        new_y = int(yOffset + mouse_y - new_mouse_y)
        if new_x < 0:
            new_x = 0
        elif new_x > self.x:
            new_x = self.x
        if new_y < 0:
            new_y = 0
        elif new_y > self.y:
            new_y = self.y

        # Set the new width, height, and position of the widget
        self.set_zoom_region(new_x, new_y, new_width, new_height)

    def updateMouseDialog(self, event):
        """
        This method continuously displays and updates the dialog box beneath the mouse cursor.

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        # Return if there is no image data
        if not self.x or not self.y:
            return
        
        # Flag to indicate mouse dialog has been launched
        self.mouse_dialog.mouse_dialog_launched = True
        
        # Get the current image size on screen
        img_width, img_height, pixel_size = self.calc_img_size_on_screen()

        # Get the current mouse position on the widget
        mouse_position = event.pos()

        # Get zoom parameters
        xOffset, yOffset, width, height = self.get_zoom_region()

        # Calculate pixel coordinates
        self.mouse_dialog.pix_x = int(mouse_position.x()//pixel_size) + xOffset
        self.mouse_dialog.pix_y = int(mouse_position.y()//pixel_size) + yOffset
        if self.mouse_dialog.pix_x >= xOffset + width:
            self.mouse_dialog.pix_x = xOffset + width - 1
        if self.mouse_dialog.pix_y >= yOffset + height:
            self.mouse_dialog.pix_y = yOffset + height - 1

        # Update the textbox location inside the widget screen as the mouse moves
        self.mouse_dialog.x_coor = mouse_position.x()+5
        self.mouse_dialog.y_coor = mouse_position.y()+10
        
        # Calculate the intensity coordinates corrected for zoom
        self.mouse_dialog.int_x = self.mouse_dialog.pix_x - xOffset
        self.mouse_dialog.int_y = self.mouse_dialog.pix_y - yOffset

        # if mouse cursor outside image window, hide mouse dialog
        if (mouse_position.y() >= (height * pixel_size)) or (mouse_position.x() >= (width * pixel_size)):
            self.mouse_dialog.mouse_dialog_hide = True
        else:
            self.mouse_dialog.mouse_dialog_hide = False

        # Set the dialog box paramters
        self.setup_mouse_textbox()

        # Flip the textbox location when approaching image widget screen boundary
        if (mouse_position.x() >= (width * pixel_size - self.mouse_dialog.max_textbox_width)):
            self.mouse_dialog.x_coor = self.mouse_dialog.x_coor - self.mouse_dialog.max_textbox_width - 20
        if (mouse_position.y() >= (height * pixel_size - 50)):
            self.mouse_dialog.y_coor = self.mouse_dialog.y_coor - 55
        
        # Set layout
        self.mouse_dialog.layout.setContentsMargins(self.mouse_dialog.x_coor, self.mouse_dialog.y_coor,int(img_width), int(img_height))
        self.mouse_dialog.layout.addWidget(self.mouse_dialog.textbox)

    def setup_mouse_textbox(self):
        """
        This method set the mouse textbox size and text within based on the image color and data type.

        """
        # if mouse cursor outside image window, hide mouse dialog
        if self.mouse_dialog.mouse_dialog_hide or self.underMouse() is False:
            self.mouse_dialog.textbox.setHidden(True)
            return
        else:
            self.mouse_dialog.textbox.setHidden(False)
        
        if self.color_mode == COLOR_MODE_MONO:
            # Get the monochromatic intensity value at the particular pixel and display in the correct format
            try:
                intensity = self.last_displayed_image.image[self.mouse_dialog.int_x][self.mouse_dialog.int_y]
                if intensity >= 999999:
                    intensity = "{:e}".format(intensity)
                self.mouse_dialog.textbox.setText(f"({self.mouse_dialog.pix_x}, {self.mouse_dialog.pix_y})\nI: {intensity}")
                # Set textbox size
                if (self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width() > self.mouse_dialog.max_textbox_width):
                    self.mouse_dialog.max_textbox_width = self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width()
                self.mouse_dialog.textbox.setFixedSize(self.mouse_dialog.max_textbox_width+5, 45)
            except Exception as e:
                logging.getLogger().error('Error displaying mouse dialog: %s', str(e))
        else:
            # Get the RGB intensity values at the particular pixel and display in the correct format
            try:
                r_intensity = self.last_displayed_image.image[self.mouse_dialog.int_x][self.mouse_dialog.int_y][0]
                if r_intensity >= 999999:
                    r_intensity = "{:e}".format(r_intensity)
                g_intensity = self.last_displayed_image.image[self.mouse_dialog.int_x][self.mouse_dialog.int_y][1]
                if g_intensity >= 999999:
                    g_intensity = "{:e}".format(g_intensity)
                b_intensity = self.last_displayed_image.image[self.mouse_dialog.int_x][self.mouse_dialog.int_y][2]
                if b_intensity >= 999999:
                    b_intensity = "{:e}".format(b_intensity)
                
                # Set textbook text and size
                if self._inputType == 'ubyteValue' or self._inputType == 'byteValue' or self._inputType == 'ushortValue' or self._inputType == 'shortValue':
                    self.mouse_dialog.textbox.setText(f"({self.mouse_dialog.pix_x}, {self.mouse_dialog.pix_y})\nR: {r_intensity}, G: {g_intensity}, B: {b_intensity}")
                    if (self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width() > self.mouse_dialog.max_textbox_width):
                        self.mouse_dialog.max_textbox_width = self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width()
                    self.mouse_dialog.textbox.setFixedSize(self.mouse_dialog.max_textbox_width+5, 45)
                else:
                    self.mouse_dialog.textbox.setText(f"({self.mouse_dialog.pix_x}, {self.mouse_dialog.pix_y})\nR: {r_intensity}")
                    if (self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width() > self.mouse_dialog.max_textbox_width):
                        self.mouse_dialog.max_textbox_width = self.mouse_dialog.textbox.fontMetrics().boundingRect(self.mouse_dialog.textbox.toPlainText()).width()
                    self.mouse_dialog.textbox.setText(f"({self.mouse_dialog.pix_x}, {self.mouse_dialog.pix_y})\nR: {r_intensity},\nG: {g_intensity},\nB: {b_intensity}")
                    self.mouse_dialog.textbox.setFixedSize(self.mouse_dialog.max_textbox_width+5, 85)
            except Exception as e:
                logging.getLogger().error('Error displaying mouse dialog: %s', str(e))


    def enable_moving_average(self, moving_average_enabled, n_moving_average_frames):
        """
        Setup moving average parameters

        :param moving_average_enabled: (bool) Enable or disable moving average
        :param n_moving_average_frames: (int) Number of frames to use for average
        """
        self.moving_average_enabled = moving_average_enabled
        self.n_moving_average_frames = n_moving_average_frames

    def set_display_queue_size(self, new_size):
        """
        Change display queue max size.

        :param int new_size: Set queue size. Valid values are >=0.
        """

        if not isinstance(new_size, int) or new_size < 0:
            return

        self.draw_queue.maxsize = new_size
        while self.draw_queue.qsize() > new_size:
            try:
                self.draw_queue.get_nowait()
            except queue.Empty:
                break

    def get_display_max_queue_size(self):
        """
        Get queue max size.
        """
        return self.draw_queue.maxsize

    def get_display_queue_size(self):
        """
        Get current display queue size.
        """
        return self.draw_queue.qsize()

    def __calculateZoomParameters(self, xminMouse, xmaxMouse, yminMouse, ymaxMouse):
        """
        This method calculate pixel offsets and sizes of the image to get zoomed image on the selected ROI.

        :param xminMouse: (Int) Min X coordinate of the selection in the display pixels. Relative to the widget.
        :param xmaxMouse: (Int) Max X coordinate of the selection in the display pixels. Relative to the widget.
        :param yminMouse: (Int) Min Y coordinate of the selection in the display pixels. Relative to the widget.
        :param ymaxMouse: (Int) Max Y coordinate of the selection in the display pixels. Relative to the widget.
        :return: (None)
        """
        # Widget sizes
        wWidth = self.width()
        wHeight = self.height()

        # Current zoom parameters
        xOffset, yOffset, width, height = self.get_zoom_region()

        # Get pixel size
        pixelSize = min (
            wWidth / width,
            wHeight / height
        )

        # Calculate new x direction zoom parameters
        xOffset += int(xminMouse / pixelSize)
        width = int((xmaxMouse - xminMouse) / pixelSize)

        # Calculate new y direction zoom parameters
        yOffset += int(yminMouse / pixelSize)
        height = int((ymaxMouse - yminMouse) / pixelSize)

        # Reset profile coordinates
        if self.image_profile_widget:
            self.image_profile_widget.setXAxisRange(xOffset,xOffset+width)
            self.image_profile_widget.setYAxisRange(yOffset,yOffset+height)

        # Write to dict
        self.set_zoom_region(xOffset, yOffset, width, height)

    def setup_profiles(self, grid_layout):
        """
        Build image_profile_widget with the specified grid layout.

        :param grid_layout: (QGridLayout) Reference to the grid layout where image
                                            widget is and the profiles should be added.
        :return: None
        """
        self.image_profile_widget = ImageProfileWidget(grid_layout)

    def set_zoom_region(self, xOffset, yOffset, width, height):
        """
        Set zoom region. If values are out of range, values will be clipped.

        :param xOffset: (int) Offset in X direction from start of the image in image pixels.
        :param yOffset: (int) Offset in Y direction from start of the image in image pixels.
        :param width: (int) Width of the image in image pixels to display.
        :param height: (int) Height of the image in image pixels to display.
        """
        if width < self.ZOOM_LENGTH_MIN:
            width = self.ZOOM_LENGTH_MIN
        if height < self.ZOOM_LENGTH_MIN:
            height = self.ZOOM_LENGTH_MIN

        if xOffset < self.x - self.ZOOM_LENGTH_MIN:
            self.__zoomDict['xoffset'] = xOffset
        else:
            self.__zoomDict['xoffset'] = self.x - self.ZOOM_LENGTH_MIN

        if yOffset < self.y - self.ZOOM_LENGTH_MIN:
            self.__zoomDict['yoffset'] = yOffset
        else:
            self.__zoomDict['yoffset'] = self.y - self.ZOOM_LENGTH_MIN

        if self.__zoomDict['xoffset'] + width <= self.x:
            self.__zoomDict['width'] = width
        else:
            self.__zoomDict['width'] = self.x - self.__zoomDict['xoffset']

        if self.__zoomDict['yoffset'] + height <= self.y:
            self.__zoomDict['height'] = height
        else:
            self.__zoomDict['height'] = self.y - self.__zoomDict['yoffset']

        self.__zoomDict['isZoom'] = True

        if self.data:
            try:
                self.display(self.data, zoomUpdate=True)
            except Exception as e:
                logging.getLogger().error('Error displaying data: %s', str(e))
    
    def pan_zoom_window(self, event):
        """
        This method allows a user to move the zoom window around the image by dragging 
        their mouse while holding down the right button

        :param event: (QMouseEvent) Parameter holding event details.
        :return: (None)
        """
        # Get the current image size on screen
        _, _, pixel_size = self.calc_img_size_on_screen()

        # Get the current mouse position on the widget
        new_mouse_position = event.pos()

        # Get zoom parameters
        xOffset, yOffset, width, height = self.get_zoom_region()

        # Calculate new offset coordinates for zoom window
        new_x = int(new_mouse_position.x()//pixel_size) - int(self.click_x//pixel_size) + xOffset
        new_y = int(new_mouse_position.y()//pixel_size) - int(self.click_y//pixel_size) + yOffset
        if new_x < 0:
            new_x = 0
        elif new_x > self.x - width:
            new_x = self.x - width
        if new_y < 0:
            new_y = 0
        elif new_y > self.y - height:
            new_y = self.y - height

        # Set the new width, height, and position of the widget
        self.set_zoom_region(new_x, new_y, width, height)

    def set_freeze(self, flag):
        """
        Freeze or unfreeze the image.

        :parama flag: (bool) True to freeze the image, False otherwise.
        :return:
        """
        self._freeze = flag
        if self.datasource is not None:
            if self._freeze:
                self.stop()
            else:
                self.start()

    def reset_zoom(self):
        """
        This method will reset __zoomDict to the default values (no zoom).

        :return: (None)
        """
        self.__zoomDict['isZoom'] = False
        self.__zoomDict['xoffset'] = 0
        self.__zoomDict['yoffset'] = 0
        self.__zoomDict['width'] = self.x
        self.__zoomDict['height'] = self.y
        if self.image_profile_widget:
            self.image_profile_widget.setXAxisRange()
            self.image_profile_widget.setYAxisRange()

        if self.data:
            try:
                self.display(self.data, zoomUpdate=True)
            except Exception as e:
                logging.getLogger().error('Error displaying data: %s', str(e))

    def is_zoomed(self):
        """
        Check if displayed image is zoomed.

        :return: (bool) True if image is zoomed, False otherwise.
        """
        return self.__zoomDict['isZoom']

    def get_zoom_region(self):
        """
        Get displayed region in image pixels.

        :return: (None or tuple)
        """
        return (
                self.__zoomDict['xoffset'],
                self.__zoomDict['yoffset'],
                self.__zoomDict['width'],
                self.__zoomDict['height'],
                )

    def calc_img_size_on_screen(self):
        """
        Calculate how big is the image on the display in pixels.

        :return: (int, int) Image width and height in pixels.
        """
        # Widget sizes
        window_width = self.width()
        window_height = self.height()

        # Current zoom parameters
        _, _, width, height = self.get_zoom_region()

        if width == 0 or height == 0:
            return

        # Get pixel size
        pixel_size = min (
            window_width / width,
            window_height / height
        )

        image_width = pixel_size * width
        image_height = pixel_size * height

        return image_width, image_height, pixel_size

    def wait(self):
        """

        :return:
        """
        self.mutex.lock()

    def signal(self):
        """

        :return:
        """
        self.mutex.unlock()

    def set_datasource(self, source):
        """
        Set data source, and start data taking

        :param source:
        :return:
        """
        if source is not None:
            self.datasource = source
        
    def __update_dimension(self, data):
        """

        :param data:
        :return:
        """

        try:
            # Get dimensions
            dims = data['dimension']
            self.dimensions = len(dims)
        

            # Get color mode
            attributes = data['attribute']
            if any(('name' in attr and attr['name'] == "ColorMode") for attr in attributes ):
                for attribute in attributes:
                    if attribute['name'] == "ColorMode":
                        self.color_mode = attribute['value'][0]['value']
            else:
                raise RuntimeError(f"NDArray does not contain ColorMode Attribute.")
        except:
            if self.dimensions == 2:
                self.color_mode = COLOR_MODE_MONO
            else:
                raise
            
        if self.dimensions == 2 and self.color_mode == COLOR_MODE_MONO:
            x = dims[0]
            y = dims[1]
            z = None
        elif self.dimensions == 3 and self.color_mode == COLOR_MODE_RGB1:
            x = dims[1]
            y = dims[2]
            z = dims[0]
        elif self.dimensions == 3 and self.color_mode == COLOR_MODE_RGB2:
            x = dims[0]
            y = dims[2]
            z = dims[1]
        elif self.dimensions == 3 and self.color_mode == COLOR_MODE_RGB3:
            x = dims[0]
            y = dims[1]
            z = dims[2]
        else:
            raise RuntimeError(f"Invalid data/color mode data. dimensions=%r, colormode=%r" % (self.dimensions, self.color_mode) )

        if (x['size'] != self.x) or (y['size'] != self.y) or (z is not None and z['size'] != self.z):
            self.x = x['size']
            self.y = y['size']
            self.z = z['size'] if z is not None else None
            self.reset_zoom()

    def set_black(self, value):
        """

        :param value:
        :return:
        """
        self._black = value

    def get_black(self):
        """

        :param value:
        :return:
        """
        return self._black

    def set_white(self, value):
        """

        :param value:
        :return:
        """
        self._white = value

    def get_white(self):
        """

        :param value:
        :return:
        """
        return self._white

    def start(self):
        """

        :return:
        """
        if not self._freeze:
            self.datasource.start(routine=self.data_callback,
                                  status_callback=self.connection_changed_signal.emit)
            
    def stop(self):
        """

        :return:
        """
        try:
            if self.datasource is not None:
                self.datasource.stop()
        except RuntimeError as e:
            logging.getLogger().debug('Error stopping source: ' + str(e))

    def set_framerate(self, value):
        """

        :param value:
        :return:
        """
        self.wait()
        self.stop()
        if value == -1:
            value = None
        self.datasource.update_framerate(value)
        if self._pref['EnableServerQueue']:
            self.datasource.update_server_queue_size(self._pref['ServerQueueSize'])
        else:
            self.datasource.update_server_queue_size(0)
        self.start()
        self.signal()
        
    def data_callback(self, data):
        """

        :param data:
        :return:
        """
        self.data = data
        self.wait()
        try:
            self.display(self.data)
        except Exception as e:
            logging.getLogger().error('Error displaying data: %s', str(e))
        finally:
            self.signal()

    def camera_changed(self, value):
        """

        :return:
        :raise RuntimeError:
        """
        self.wait()
        self._agc = False
        self._lastTimestamp = None
        try:
            self.datasource.update_device(value, test_connection=False)
        except pva.PvaException as e:
            # TODO wrap PvaException from pvaPy in a better way for other interface
            # pvAccess connection error
            # release mutex lock
            self.signal()
            # NOTE: data source is already stopped on error. calling stop here will override
            # the connection status
            #self.stop()
            raise RuntimeError(str(e))

        try:
            self.set_scaling()
            self.start()
            self.signal()
        except ValueError as e:
            self.stop()
            self.signal()
            raise e

    def get_preferences(self):
        """
        :return: (dict) Preferences dictionary.
        """
        return self._pref

    def set_preferences(self, preferences):
        """
        Set preferences.

        :param preferences: (dist) Preference dict.
        """
        self._pref.update(preferences)

    def enable_auto_white(self):
        """
        Enable auto white calibration for image

        :return:
        """
        self._agc = True

    def set_BlackLimitsCallback(self, function):
        """

        :return:
        """
        self.updateGuiBlackLimits = function

    def set_WhiteLimitsCallback(self, function):
        """

        :return:
        """
        self.updateGuiWhiteLimits = function

    def set_getBlackWhiteLimits(self, getBlackLimitsFunction, getWhiteLimitsFunction):
        """

        :return:
        """
        self.getGuiBlackLimits = getBlackLimitsFunction
        self.getGuiWhiteLimits = getWhiteLimitsFunction

    def set_BlackCallback(self, function):
        """

        :return:
        """
        self.updateGuiBlack = function

    def set_WhiteCallback(self, function):
        """

        :return:
        """
        self.updateGuiWhite = function

    def set_scaling(self, scale=None):
        """

        :param scale:
        :return:
        """
        if scale is None:
            if self.x is None:
                self._scaling = 1.0
            else:
                self._scaling = self.width()/self.x
        else:
            self._scaling = scale

    def display(self, data, zoomUpdate = False):
        """
        Display and update image using passed data.

        :param data: Channel PV data.
        :param zoomUpdate: (bool) Must be False to display new image. True if method
                            is called to update the current displayed image with new
                            zoom.
        :return:
        """
        # Count received frames
        current_array_id = data['uniqueId']
        if not zoomUpdate:
            if self.__last_array_id == current_array_id:
                # Frame has been seen already
                return
            self.frames_received += 1

        # Check if the queue is full
        if self.draw_queue.full() and not zoomUpdate:
            return

        # Update dimensions
        if not zoomUpdate:
            self.__update_dimension(data)

        # Determinate input data type
        data_types = data['value'][0].keys()
        assert len(list(data_types)) == 1
        inputType = list(data_types)[0]

        # Check if the input is valid
        if not inputType in self.dataTypesDict or not self.color_mode in COLOR_MODES:
            self._isInputValid = False
            raise RuntimeError('No recognized image data received.')
        else:
            self._isInputValid = True

        # Lookup required infomartion for the current input type
        imgArray = data['value'][0][inputType]
        sz = imgArray.nbytes
        codecName = data['codec']['name']

        # Check if image size is zero
        if imgArray.size == 0:
            raise RuntimeError('Image size cannot be zero.')
        
        if codecName:
            uncompressedSize = data['uncompressedSize']
            uncompressedType = data['codec.parameters'][0]['value']
            decompress = ImageCompressionUtility.get_decompressor(codecName)
            imgArray = decompress(imgArray, uncompressedType, uncompressedSize)
            inputType = ImageCompressionUtility.get_ntnda_data_type(uncompressedType)

        # Calculate average image if needed
        moving_average_enabled = self.moving_average_enabled
        n_moving_average_frames = self.n_moving_average_frames
        if moving_average_enabled and n_moving_average_frames > 1:
            dtype = imgArray.dtype
            image = imgArray.astype(np.float64)
            self.image_list.insert(0,image)
            n_images = min(len(self.image_list),n_moving_average_frames)
            self.image_list = self.image_list[:n_images]
            average_image = sum(self.image_list)/n_images
            average_image = average_image.astype(dtype)
            imgArray = average_image
        else:
           self.image_list = []

        maxVal = self.dataTypesDict[inputType]['maxVal']
        minVal = self.dataTypesDict[inputType]['minVal']
        npdt = self.dataTypesDict[inputType]['npdt']
        embeddedDataLen = self._pref['EmbeddedDataLen']

        # Configure GUI for the correct type
        if inputType != self._inputType:
            self.configureGuiLimits(inputType)

        self._inputType = inputType

        # Handle timestamp
        ts = data['timeStamp']
        if ts == self._lastTimestamp and not zoomUpdate:
            # if timstamp hasn't changed, its not a new image
            return
        self._lastTimestamp = ts

        # Update numbers of MBytes received (bytes => kBytes => MBytes conversion)
        if not zoomUpdate:
            self.MB_received += sz/1000.0/1000.0

        # Get image statistics
        if maxVal != self.maxVal:
            self.maxVal = maxVal
        if minVal != self.minVal:
            self.minVal = minVal

        # Resize to get a 2D array from 1D data structure
        if self.x == 0 or self.y == 0 or self.color_mode is None:
            # return
            raise RuntimeError("Image dimension not initialized")
        img = transcode_image(imgArray, self.color_mode, self.x, self.y, self.z)

        # Count dead pixels
        if self._pref['DPXEnabled']:
            current_img_dead_pixels = (imgArray[embeddedDataLen:] > self._pref["DPXLimit"]).sum()
        else:
            current_img_dead_pixels = 0
        self.dead_pixels.append(current_img_dead_pixels)
        self.dead_pixels = self.dead_pixels[-3:]

        # Record max/min pixel values on ROI
        if img.ndim == 2:
            self._max = imgArray[embeddedDataLen:].max()
            self._min = imgArray[embeddedDataLen:].min()
        elif img.ndim == 3:
            self._max = (
                img[:,:,0].flatten()[embeddedDataLen:].max(),
                img[:,:,1].flatten()[embeddedDataLen:].max(),
                img[:,:,2].flatten()[embeddedDataLen:].max(),
            )
            self._min = (
                img[:,:,0].flatten()[embeddedDataLen:].min(),
                img[:,:,1].flatten()[embeddedDataLen:].min(),
                img[:,:,2].flatten()[embeddedDataLen:].min(),
            )
        else:
            raise RuntimeError("Invalid number of dimensions image")


        # Handle image transformation
        if self.is_zoomed():
            xoffset, yoffset, width, height = self.get_zoom_region()
            endx = xoffset + width
            endy = yoffset + height
            if self.color_mode == COLOR_MODE_MONO:
                img = img[yoffset:endy, xoffset:endx]
            else:
                img = img[yoffset:endy, xoffset:endx, :]

        # Count dead pixels on ROI
        if self._pref['DPXEnabled']:
            current_img_dead_pixels = (imgArray > self._pref["DPXLimit"]).sum()
        else:
            current_img_dead_pixels = 0
        self.dead_pixels_roi.append(current_img_dead_pixels)
        self.dead_pixels_roi = self.dead_pixels_roi[-3:]

        # Record max/min pixel values on ROI
        if img.ndim == 2:
            self._maxRoi = img.flatten().max()
            self._minRoi = img.flatten().min()
        elif img.ndim == 3:
            self._maxRoi = (
                img[:,:,0].flatten().max(),
                img[:,:,1].flatten().max(),
                img[:,:,2].flatten().max(),
            )
            self._minRoi = (
                img[:,:,0].flatten().min(),
                img[:,:,1].flatten().min(),
                img[:,:,2].flatten().min(),
            )
        else:
            raise RuntimeError("Invalid number of dimensions image")

        # Auto calculate black/white levels
        if self._agc:
            black, white = np.percentile(img, [0.01, 99.99])

            if self.getGuiBlackLimits and self.getGuiWhiteLimits:
                blackMin, blackMax = self.getGuiBlackLimits()
                whiteMin, whiteMax = self.getGuiWhiteLimits()
                black = np.clip(black, blackMin, blackMax)
                white = np.clip(white, whiteMin, whiteMax)

            self.set_black(black)
            if self.updateGuiBlack is not None:
                self.updateGuiBlack(self.get_black())
            self.set_white(white)
            if self.updateGuiWhite is not None:
                self.updateGuiWhite(self.get_white())

            self._agc = False

        # Flip and rotate the image for 90 degrees (TODO: add description why)
        img = np.rot90(np.fliplr(img)).astype(npdt)

        self.calculate_profiles(img)

        # Missed frames
        if self.__last_array_id is not None and zoomUpdate == False:
            # Make sure IOC did not restart and reset counter
            frames_missed = current_array_id - self.__last_array_id - 1
            if frames_missed > 0:
                self.frames_missed += frames_missed
        self.__last_array_id = current_array_id

        image = Image(current_array_id,
                      not zoomUpdate,
                      img.copy(),
                      self.get_black(),
                      self.get_white())

        # Put the image on the queue and emit the signal
        self.draw_queue.put(image)
        self._set_image_signal.emit()

    def configureGuiLimits(self, dataType):
        """
        Configure the limits for the black and white settings on the GUI.

        :param dataType: (String) String representation of the incoming datatype. (E.g. 'ubyteValue')
        """
        blackMin = self.dataTypesDict[dataType]['minVal']
        blackMax = self.dataTypesDict[dataType]['maxVal']
        whiteMin = self.dataTypesDict[dataType]['minVal']
        whiteMax = self.dataTypesDict[dataType]['maxVal']

        if self.updateGuiBlackLimits is not None:
            self.updateGuiBlackLimits(blackMin, blackMax)
        if self.updateGuiWhiteLimits is not None:
            self.updateGuiWhiteLimits(whiteMin, whiteMax)

    def calculate_profiles(self, image):
        """
        Handle profile calculation.

        :param image: (nparray) Image as numpy array.
        """
        if self.image_profile_widget is not None:
            self.image_profile_widget.set_image_data(image, self.color_mode)

    def _set_image_signal_callback(self):
        """
        Update image on the widget.

        :return: None
        """
        image = self.draw_queue.get()
        levels = [image.black, image.white]
        self.setImage(image.image, levels = levels)
        if self.image_profile_widget is not None:
            img_width, img_height, _ = self.calc_img_size_on_screen()
            self.image_profile_widget.plot(img_width, img_height)
        if image.new:
            self.frames_displayed += 1
        self.last_displayed_image = image

        # If the mouse dialog has been launched, update the intensity value with every image update
        if (self.mouse_dialog.mouse_dialog_launched):
            # Set the dialog box paramters
            self.setup_mouse_textbox()

            # Set layout
            self.mouse_dialog.layout.addWidget(self.mouse_dialog.textbox)



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
from configparser import ConfigParser
from c2dataviewer.control.striptoolconfig import StripToolConfigure
from c2dataviewer.control.pvconfig import PvConfig
from c2dataviewer.model import make_protocol

# Global protocol constants for testing
PROTO_CA = make_protocol('ca')
PROTO_PVA = make_protocol('pva')


class TestStriptoolConfig(unittest.TestCase):
    def setUp(self):
        """Reset color index before each test."""
        PvConfig.color_index = 0

    def test_config(self):
        raw = """
[STRIPTOOL]
DefaultProtocol = ca
Chan1.PV = Foo:Bar
Chan2.PV = pva://Bar:Baz
Chan1.Color = #000000
Chan2.Color = #0000FF
"""
        parser = ConfigParser()
        parser.read_string(raw)
        cfg = StripToolConfigure(parser)
        expected = {
            'Foo:Bar': PvConfig('Foo:Bar', '#000000', 'ca'),
            'Bar:Baz': PvConfig('Bar:Baz', '#0000FF', 'pva')
        }

        self.assertEqual(len(cfg.pvs), len(expected))

        for pv in cfg.pvs.values():
            self.assertTrue(pv.pvname in expected)
            e = expected[pv.pvname]
            self.assertEqual(e.pvname, pv.pvname)
            self.assertEqual(e.color, pv.color)
            self.assertEqual(e.proto, pv.proto)
            del expected[pv.pvname]

    def get_display_val(self, data, val):
        for child in data['children']:
            if child['name'] == val:
                ret_val = child['value']
                return ret_val

    # ------------------------------------------------------------------
    # Display configuration tests
    # ------------------------------------------------------------------
    def test_display_configuration(self):
        """
        Comprehensive test for display configuration including:
        - Autoscale precedence (STRIPTOOL section overrides DISPLAY section)
        - All settings with valid values
        - Default values when settings absent (StripTool default autoscale is True)
        - Bad/invalid input handling with fallback to defaults
        """
        # Test 1: Autoscale precedence - STRIPTOOL section overrides DISPLAY
        raw1 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        AUTOSCALE=True

        [DISPLAY]
        AUTOSCALE=False
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))

        # Test 2: Autoscale from DISPLAY section when not in STRIPTOOL
        raw2 = """
        [STRIPTOOL]
        DefaultProtocol = ca

        [DISPLAY]
        AUTOSCALE=False
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        self.assertFalse(self.get_display_val(data=display, val='Autoscale'))

        # Test 3: Default autoscale (True for StripTool)
        raw3 = """
        [STRIPTOOL]
        DefaultProtocol = ca

        [DISPLAY]
        AVERAGE=1
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        self.assertTrue(self.get_display_val(data=display, val='Autoscale'))

        # Test 4: All settings with valid values
        raw4 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        AUTOSCALE=False

        [DISPLAY]
        MODE=psd
        FFT_FILTER=hamming
        AVERAGE=15
        SINGLE_AXIS=False
        HISTOGRAM=True
        N_BIN=512
        REFRESH=250
        """
        parser = ConfigParser()
        parser.read_string(raw4)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        self.assertEqual(self.get_display_val(display, 'Mode'), 'psd')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'hamming')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 15)
        self.assertFalse(self.get_display_val(display, 'Autoscale'))  # STRIPTOOL precedence
        self.assertFalse(self.get_display_val(display, 'Single axis'))
        self.assertTrue(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 512)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.25)  # 250ms -> 0.25s

        # Test 5: Default values when no DISPLAY section
        raw5 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        """
        parser = ConfigParser()
        parser.read_string(raw5)
        configure = StripToolConfigure(parser)
        display = configure.assemble_display(section=None)
        self.assertEqual(self.get_display_val(display, 'Mode'), 'normal')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'none')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 1)
        self.assertTrue(self.get_display_val(display, 'Autoscale'))
        self.assertTrue(self.get_display_val(display, 'Single axis'))
        self.assertFalse(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 100)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.1)

        # Test 6: Bad/invalid input falls back to defaults
        raw6 = """
        [STRIPTOOL]
        DefaultProtocol = ca

        [DISPLAY]
        MODE=invalidmode
        FFT_FILTER=bogusfilter
        AVERAGE=-10
        SINGLE_AXIS=maybe
        HISTOGRAM=maybe
        N_BIN=badint
        REFRESH=badfloat
        AUTOSCALE=maybe
        """
        parser = ConfigParser()
        parser.read_string(raw6)
        configure = StripToolConfigure(parser)
        section = parser["DISPLAY"]
        display = configure.assemble_display(section=section)
        self.assertEqual(self.get_display_val(display, 'Mode'), 'normal')
        self.assertEqual(self.get_display_val(display, 'FFT filter'), 'none')
        self.assertEqual(self.get_display_val(display, 'Exp moving avg'), 1)
        self.assertTrue(self.get_display_val(display, 'Single axis'))
        self.assertFalse(self.get_display_val(display, 'Histogram'))
        self.assertEqual(self.get_display_val(display, 'Num Bins'), 100)
        self.assertEqual(self.get_display_val(display, 'Refresh'), 0.1)
        self.assertTrue(self.get_display_val(display, 'Autoscale'))

    # ------------------------------------------------------------------
    # Channel/PV configuration tests
    # ------------------------------------------------------------------
    def test_channel_configuration(self):
        """
        Comprehensive test for channel/PV configuration including:
        - Basic PV names and colors
        - Protocol specification and defaults
        - Default color assignment
        - Case-insensitive channel names
        - Multiple channels
        - Special characters in PV names
        """
        # Test 1: Basic channel configuration
        raw1 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV1
        Chan1.Color = #FF0000
        Chan2.PV = TEST:PV2
        Chan2.Color = #00FF00
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 2)
        self.assertIn('TEST:PV1', cfg.pvs)
        self.assertIn('TEST:PV2', cfg.pvs)
        pv1 = cfg.pvs['TEST:PV1']
        self.assertEqual(pv1.pvname, 'TEST:PV1')
        self.assertEqual(pv1.color, '#FF0000')
        self.assertEqual(pv1.proto, PROTO_CA)
        pv2 = cfg.pvs['TEST:PV2']
        self.assertEqual(pv2.pvname, 'TEST:PV2')
        self.assertEqual(pv2.color, '#00FF00')
        self.assertEqual(pv2.proto, PROTO_CA)

        # Test 2: Protocol override in PV string
        raw2 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = ca://TEST:PV1
        Chan2.PV = pva://TEST:PV2
        Chan3.PV = TEST:PV3
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 3)
        self.assertEqual(cfg.pvs['TEST:PV1'].proto, PROTO_CA)
        self.assertEqual(cfg.pvs['TEST:PV2'].proto, PROTO_PVA)
        self.assertEqual(cfg.pvs['TEST:PV3'].proto, PROTO_CA)

        # Test 3: Default protocol pva
        raw3 = """
        [STRIPTOOL]
        DefaultProtocol = pva
        Chan1.PV = TEST:PV1
        Chan2.PV = ca://TEST:PV2
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        cfg = StripToolConfigure(parser)
        self.assertEqual(cfg.pvs['TEST:PV1'].proto, PROTO_PVA)
        self.assertEqual(cfg.pvs['TEST:PV2'].proto, PROTO_CA)

        # Test 4: Default color assignment
        PvConfig.color_index = 0  # Reset color index
        raw4 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV1
        Chan2.PV = TEST:PV2
        Chan2.Color = #FFFFFF
        """
        parser = ConfigParser()
        parser.read_string(raw4)
        cfg = StripToolConfigure(parser)
        self.assertEqual(cfg.pvs['TEST:PV1'].color, '#FFFF00')  # Default color
        self.assertEqual(cfg.pvs['TEST:PV2'].color, '#FFFFFF')  # Explicit color

        # Test 5: Case-insensitive channel names
        raw5 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        CHAN1.PV = TEST:PV1
        Chan2.pv = TEST:PV2
        chan3.PV = TEST:PV3
        """
        parser = ConfigParser()
        parser.read_string(raw5)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 3)
        self.assertIn('TEST:PV1', cfg.pvs)
        self.assertIn('TEST:PV2', cfg.pvs)
        self.assertIn('TEST:PV3', cfg.pvs)

        # Test 6: Multiple properties
        raw6 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = pva://TEST:MOTOR1
        Chan1.Color = #123456
        """
        parser = ConfigParser()
        parser.read_string(raw6)
        cfg = StripToolConfigure(parser)
        pv = cfg.pvs['TEST:MOTOR1']
        self.assertEqual(pv.pvname, 'TEST:MOTOR1')
        self.assertEqual(pv.color, '#123456')
        self.assertEqual(pv.proto, PROTO_PVA)

        # Test 7: Empty section
        raw7 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        """
        parser = ConfigParser()
        parser.read_string(raw7)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 0)
        self.assertEqual(cfg.default_proto, 'ca')

        # Test 8: Many channels
        raw8 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV1
        Chan2.PV = TEST:PV2
        Chan3.PV = TEST:PV3
        Chan4.PV = TEST:PV4
        Chan5.PV = TEST:PV5
        Chan6.PV = TEST:PV6
        Chan7.PV = TEST:PV7
        Chan8.PV = TEST:PV8
        """
        parser = ConfigParser()
        parser.read_string(raw8)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 8)
        for i in range(1, 9):
            self.assertIn(f'TEST:PV{i}', cfg.pvs)

        # Test 9: Duplicate PVs (last one wins)
        raw9 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:SAME
        Chan1.Color = #FF0000
        Chan2.PV = TEST:SAME
        Chan2.Color = #00FF00
        """
        parser = ConfigParser()
        parser.read_string(raw9)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 1)
        pv = cfg.pvs['TEST:SAME']
        self.assertEqual(pv.color, '#00FF00')

        # Test 10: Special characters in PV names
        raw10 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV-WITH-DASHES
        Chan2.PV = TEST:PV_WITH_UNDERSCORES
        Chan3.PV = TEST:PV.WITH.DOTS
        Chan4.PV = TEST:PV:COLONS:HERE
        """
        parser = ConfigParser()
        parser.read_string(raw10)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 4)
        self.assertIn('TEST:PV-WITH-DASHES', cfg.pvs)
        self.assertIn('TEST:PV_WITH_UNDERSCORES', cfg.pvs)
        self.assertIn('TEST:PV.WITH.DOTS', cfg.pvs)
        self.assertIn('TEST:PV:COLONS:HERE', cfg.pvs)

    # ------------------------------------------------------------------
    # Error handling and edge cases
    # ------------------------------------------------------------------
    def test_error_handling(self):
        """
        Comprehensive test for error handling and edge cases including:
        - Missing STRIPTOOL section
        - Missing DefaultProtocol
        - Invalid color formats
        - Channel with only color but no PV
        - Protocol string format variations
        """
        # Test 1: Missing STRIPTOOL section
        raw1 = """
        [DISPLAY]
        MODE=normal
        """
        parser = ConfigParser()
        parser.read_string(raw1)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 0)
        self.assertIsNone(cfg.default_proto)

        # Test 2: Missing DefaultProtocol (should default to 'ca')
        raw2 = """
        [STRIPTOOL]
        Chan1.PV = TEST:PV1
        """
        parser = ConfigParser()
        parser.read_string(raw2)
        cfg = StripToolConfigure(parser)
        self.assertEqual(cfg.default_proto, 'ca')
        self.assertEqual(cfg.pvs['TEST:PV1'].proto, PROTO_CA)

        # Test 3: Invalid color formats (stored as-is)
        raw3 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV1
        Chan1.Color = red
        Chan2.PV = TEST:PV2
        Chan2.Color = 12345
        """
        parser = ConfigParser()
        parser.read_string(raw3)
        cfg = StripToolConfigure(parser)
        self.assertEqual(cfg.pvs['TEST:PV1'].color, 'red')
        self.assertEqual(cfg.pvs['TEST:PV2'].color, '12345')

        # Test 4: Channel with only color but no PV (ignored)
        raw4 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.Color = #FF0000
        Chan2.PV = TEST:PV2
        """
        parser = ConfigParser()
        parser.read_string(raw4)
        cfg = StripToolConfigure(parser)
        self.assertEqual(len(cfg.pvs), 1)
        self.assertIn('TEST:PV2', cfg.pvs)

        # Test 5: Protocol string case variations
        raw5 = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = CA://TEST:PV1
        Chan2.PV = PVA://TEST:PV2
        Chan3.PV = pva://TEST:PV3
        """
        parser = ConfigParser()
        parser.read_string(raw5)
        cfg = StripToolConfigure(parser)
        self.assertIn('TEST:PV1', cfg.pvs)
        self.assertIn('TEST:PV2', cfg.pvs)
        self.assertIn('TEST:PV3', cfg.pvs)

    def test_acquisition_sample_mode(self):
        """
        Test that acquisition includes Sample Mode for striptool.
        """
        raw = """
        [STRIPTOOL]
        DefaultProtocol = ca
        """
        parser = ConfigParser()
        parser.read_string(raw)
        cfg = StripToolConfigure(parser)

        acquisition = cfg.assemble_acquisition()

        # Check that Sample Mode is present
        sample_mode = None
        for child in acquisition['children']:
            if child['name'] == 'Sample Mode':
                sample_mode = child
                break

        self.assertIsNotNone(sample_mode)
        self.assertEqual(sample_mode['type'], 'bool')
        self.assertTrue(sample_mode['value'])


if __name__ == '__main__':
    unittest.main()



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import numpy as np
import pyqtgraph
import pvaccess as pva
from ..model import ConnectionState
import math
from .config import Scope
from PyQt5 import QtWidgets

class ScopeControllerBase:
    def __init__(self, widget, model, parameters, warning, channels=[], **kwargs):
        self._win = widget
        self.model = model
        self.parameters = parameters
        self.data = None
        # refresh frequency: every 100 ms by default
        self.refresh = 100

        self.plotting_started = False

        self.timer = pyqtgraph.QtCore.QTimer()
        self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
        self._win.graphicsWidget.set_autoscale(parameters.child('Display', 'Autoscale').value())
        self._win.graphicsWidget.set_model(self.model)

        self._warning = warning
        self._warning.warningConfirmButton.clicked.connect(lambda: self.accept_warning())

        self.arrays = np.array([])
        self.lastArrays = 0

        self.default_trigger = None
        self.trigger_is_monitor = False
        self.trigger_auto_scale = False

        self._win.graphicsWidget.set_histogram(parameters.child('Display', 'Histogram').value())
        single_axis = parameters.child('Display', 'Single axis').value()
        self._win.graphicsWidget.setup_plot(channels=channels, single_axis=single_axis)

        display_mode = parameters.child('Display', 'Mode').value()
        if(display_mode != "normal"):
            self.set_display_mode(display_mode)
            
        # timer to update status with statistics data
        self.status_timer = pyqtgraph.QtCore.QTimer()
        self.status_timer.timeout.connect(self.update_status)
        self.status_timer.start(1000)

        # Setup save configuration button
        self._win.saveConfigButton.clicked.connect(self._on_save_config_action)


    def set_display_mode(self, val):
        self._win.graphicsWidget.set_display_mode(val)
        self._win.graphicsWidget.setup_plot()
        # Disable multiaxis for the FFT-type modes. As at the time of the
        # writing this code pyqtgraph does not support logarithmic scale in multiaxis configuration.
        if self._win.graphicsWidget.display_mode.is_fft():
            self.parameters.child("Display").child("Single axis").setReadonly()
        else:
            self.parameters.child("Display").child("Single axis").setWritable()

    def set_trigger_pv(self, pvname):
        if self._win.graphicsWidget.plotting_started:
            self.notify_warning("Stop plotting first before changing trigger PV")
            return
                
        pvname = pvname.strip()
        if pvname == '':
            return
        
        proto = 'ca'
        name = pvname
        
        if "://" in name:
            proto, name = pvname.split("://")

        try:
            pvfields = self.model.update_trigger(name, proto=proto.lower())
            child = self.parameters.child("Trigger").child("Time Field")
            if proto == 'ca':
                child.hide()
                self._win.graphicsWidget.trigger.trigger_time_field = 'timeStamp'
            else:
                self._win.graphicsWidget.trigger.trigger_time_field = None
                pvfields.insert(0, 'None')
                child.show()
                # Get current value before updating limits
                current_value = child.value()
                child.setLimits(pvfields)
                # Restore value if it's valid, otherwise set to None
                if current_value in pvfields:
                    child.setValue(current_value)
                else:
                    child.setValue('None')                                
                                
        except Exception as e:
            self.notify_warning("Channel {}://{} timed out. \n{}".format(proto, name, repr(e)))


    def default_config(self, **kwargs):
        buffer_unit = kwargs.get('buffer_unit', 'Samples')
            
        max_length = self.parameters.child("Acquisition").child("Buffer (%s)" % buffer_unit).value()
        if max_length and buffer_unit == 'Samples':
            self.update_buffer_samples(int(max_length))
            
        self._win.graphicsWidget.set_binning(self.parameters.child("Display").child("Num Bins").value())

        refresh = self.parameters.child("Display").child("Refresh").value()
        if refresh:
            self.set_freshrate(refresh)

        try:
            trigger_pv = self.parameters.child("Trigger").child("PV").value()
            self.set_trigger_pv(trigger_pv)

            trigger_mode = self.parameters.child("Trigger").child("Mode").value()
            self.set_trigger_mode(trigger_mode)
        
            self.trigger_auto_scale = self.parameters.child("Trigger").child("Autoscale Buffer").value()
        except:
            pass

        
    def update_buffer_samples(self, size):
        """
        Sets number of samples in buffer
        
        :param size  size of buffer in number of samples
        """
        self._win.graphicsWidget.update_buffer(size)
        self.parameters.child("Acquisition").child("Buffer (Samples)").setValue(self._win.graphicsWidget.max_length)
            
    def parameter_change(self, params, changes):
        """
        
        :param params:
        :param changes:
        :return:
        """
        for param, change, data in changes:
            if change == "value":
                path = self.parameters.childPath(param)
                if path is not None:
                    childName = '.'.join(path)
                else:
                    childName = param.name()

                if childName == "Trigger.PV":
                    if data is None:
                        return
                    self.set_trigger_pv(data)
                elif childName == "Trigger.Mode":
                    self.set_trigger_mode(data)
                elif childName == "Trigger.Threshold":
                    self._win.graphicsWidget.trigger.trigger_level = data
                elif childName == "Trigger.Data Time Field":
                    self._win.graphicsWidget.trigger.data_time_field = data
                elif childName == "Trigger.Time Field":
                    self._win.graphicsWidget.trigger.trigger_time_field = data
                elif childName == "Trigger.Autoscale Buffer":
                    self.trigger_auto_scale = data
                elif childName == "Acquisition.Buffer (Samples)":
                    self._win.graphicsWidget.update_buffer(data)
                elif childName == "Acquisition.Freeze":
                    self._win.graphicsWidget.is_freeze = data
                elif childName == "Display.Mode":
                    self.set_display_mode(data)
                elif childName == "Display.FFT filter":
                    self._win.graphicsWidget.set_fft_filter(data)
                elif childName == "Display.Exp moving avg":
                    self._win.graphicsWidget.set_average(data)
                elif childName == "Display.Autoscale":
                    self._win.graphicsWidget.set_autoscale(data)
                elif childName == "Display.Single axis":
                    self._win.graphicsWidget.setup_plot(single_axis=data)
                elif childName == "Display.Histogram":
                    self._win.graphicsWidget.set_histogram(data)
                elif childName == "Display.Num Bins":
                    self._win.graphicsWidget.set_binning(data)
                elif childName == "Display.Refresh":
                    self.set_freshrate(data)

    def set_freshrate(self, value):
        """
        Set time to refresh
            
        :param value: time interval to plot, in second
        :return:
        """
        plotting_started = self._win.graphicsWidget.plotting_started
        self.stop_plotting()
        self.refresh = value*1000.0
        if plotting_started:
            self.start_plotting()


    def start_plotting(self):
        """
            
        :return:
        """

        try:
            self.timer.timeout.disconnect()
            self.timer.stop()
        except Exception:
            pass

        self.stop_trigger()
        
        # Setup free run plotting
        if not self._win.graphicsWidget.trigger_mode():
            self.timer.timeout.connect(self._win.graphicsWidget.update_drawing)
            self.timer.start(int(self.refresh))
        else:
            self.start_trigger()

        self._win.graphicsWidget.notify_plotting_started(True)
        
    def stop_plotting(self):
        """

        :return:
        """
        self._win.graphicsWidget.notify_plotting_started(False)

        self.timer.stop()
        self.stop_trigger()
                    
    def set_trigger_mode(self, value):
        """
        Set trigger mode.
    
        :param value:
        :return:
        """
        trigger_mode = value != 'none'
        if trigger_mode != self._win.graphicsWidget.trigger_mode():
            if self._win.graphicsWidget.plotting_started:
                action = 'on' if trigger_mode else 'off'
                self.notify_warning('Stop plotting first before turning trigger %s' % (action))
                return
            
        self._win.graphicsWidget.set_trigger_mode(trigger_mode)
            
        trigger_type = None
        if trigger_mode:
            self._win.graphicsWidget.trigger.set_trigger_type(value)
            
    def start_trigger(self):
        """
        Process to start DAQ in trigger mode
            
        :return:
        """
        if not self.trigger_is_monitor:
            if self.model.trigger is None:
                raise Exception('Trigger PV is not set or is invalid')
            

            
            self.model.start_trigger(self._win.graphicsWidget.trigger.data_callback)
            self.trigger_is_monitor = True

    def stop_trigger(self):
        """
        Stop trigger mode
        :return:
        """
        self.trigger_is_monitor = False
        self.model.stop_trigger()
        
    def accept_warning(self):
        """
            
        :return:
        """
        self._warning.close()

    def notify_warning(self, msg):
        #close previous warning
        self.accept_warning()
            
        self._warning.warningTextBrowse.setText(msg)
        self._warning.show()

        
    def update_status(self):
        """
        Update statistics status.
            
        :return:
        """
        # Update display
        single_axis_child = self.parameters.child("Display", "Single axis")
        if single_axis_child.value() != self._win.graphicsWidget.single_axis:
            single_axis_child.setValue(self._win.graphicsWidget.single_axis)

        # Update statistics
        with self._win._proc.oneshot():
            cpu = self._win._proc.cpu_percent(None)

        # TODO algorithm to calculate Array/sec
        arraysReceived = self._win.graphicsWidget.arraysReceived
        n = arraysReceived - self.lastArrays
        self.lastArrays = arraysReceived
        self.arrays = np.append(self.arrays, n)[-10:]

        for q in self.parameters.child("Statistics").children():
            if q.name() == 'CPU':
                q.setValue(cpu)
            elif q.name() == 'Lost Arrays':
                q.setValue(self._win.graphicsWidget.lostArrays)
            elif q.name() == 'Tot. Arrays':
                q.setValue(self._win.graphicsWidget.arraysReceived)
            elif q.name() == 'Arrays/Sec':
                q.setValue(self.arrays.mean())
            elif q.name() == 'Bytes/Sec':
                q.setValue(self.arrays.mean() * self._win.graphicsWidget.data_size)
            elif q.name() == 'Rate':
                q.setValue(self._win.graphicsWidget.fps)

        try:
            for q in self.parameters.child("Trigger").children():
                if q.name() == "Trig Status":
                    stat_str = 'Disconnected'
                    if self.trigger_is_monitor:
                        stat_str = self._win.graphicsWidget.trigger.status()
                    q.setValue(stat_str)
                elif q.name() == "Trig Value":
                    q.setValue(str(self._win.graphicsWidget.trigger.trigger_value))
        except:
            pass

        #handle any auto-adjustments
        if self.trigger_auto_scale and self._win.graphicsWidget.trigger_mode():
            if self._win.graphicsWidget.trigger.missed_triggers > 0:
                newsize = self._win.graphicsWidget.trigger.missed_adjust_buffer_size

                # Round up to 3 10's places
                # ex. 12345 would be rounded to 12400
                precision = 3
                exp = math.ceil(math.log10(newsize))
                roundunit = 10**max(exp - precision, 0)
                newsize = math.ceil(newsize / roundunit) * roundunit

                self.update_buffer_samples(newsize)

    def serialize_scope_config(self, serializer):
        """
        Serialize scope base configuration to a Serializer instance.
        This includes Display, Acquisition, and Trigger settings.

        :param serializer: config.Serializer instance to write to
        """
        # Serialize Display settings
        display_mode = self.parameters.child("Display", "Mode").value()
        serializer.set(Scope.DISPLAY_MODE, display_mode)

        fft_filter = self.parameters.child("Display", "FFT filter").value()
        serializer.set(Scope.FFT_FILTER, fft_filter)

        average = self.parameters.child("Display", "Exp moving avg").value()
        serializer.set(Scope.AVERAGE, int(average))

        autoscale = self.parameters.child("Display", "Autoscale").value()
        serializer.set(Scope.AUTOSCALE, autoscale)

        single_axis = self.parameters.child("Display", "Single axis").value()
        serializer.set(Scope.SINGLE_AXIS, single_axis)

        histogram = self.parameters.child("Display", "Histogram").value()
        serializer.set(Scope.HISTOGRAM, histogram)

        n_bin = self.parameters.child("Display", "Num Bins").value()
        serializer.set(Scope.NBIN, int(n_bin))

        refresh = self.parameters.child("Display", "Refresh").value()
        # Convert from seconds to milliseconds
        serializer.set(Scope.REFRESH, int(refresh * 1000))

        # Serialize Acquisition settings
        # Try to get buffer unit (scope-specific)
        try:
            buffer_unit = self.parameters.child("Acquisition", "Buffer Unit").value()
            if buffer_unit:
                serializer.set(Scope.BUFFER_UNIT, buffer_unit)
                # Get buffer size based on buffer unit
                buffer_size = self.parameters.child("Acquisition", f"Buffer ({buffer_unit})").value()
                if buffer_size:
                    serializer.set(Scope.BUFFER, int(buffer_size))
        except:
            # If Buffer Unit doesn't exist, fall back to Buffer (Samples)
            try:
                buffer_samples = self.parameters.child("Acquisition", "Buffer (Samples)").value()
                if buffer_samples:
                    serializer.set(Scope.BUFFER, int(buffer_samples))
            except:
                pass

        # freeze = self.parameters.child("Acquisition", "Freeze").value()
        # Freeze is not in schema, but we can document it if needed

        # Serialize Trigger settings
        try:
            trigger_pv = self.parameters.child("Trigger", "PV").value()
            if trigger_pv:
                serializer.set(Scope.TRIGGER, trigger_pv)

            trigger_mode = self.parameters.child("Trigger", "Mode").value()
            serializer.set(Scope.TRIGGER_MODE, trigger_mode)

            threshold = self.parameters.child("Trigger", "Threshold").value()
            serializer.set(Scope.TRIGGER_THRESHOLD, threshold)

            autoscale_buffer = self.parameters.child("Trigger", "Autoscale Buffer").value()
            serializer.set(Scope.TRIGGER_AUTOSCALE_BUFFER, autoscale_buffer)

            time_field = self.parameters.child("Trigger", "Time Field").value()
            if time_field and time_field != "None":
                serializer.set(Scope.TRIGGER_TIME_FIELD, time_field)

            data_time_field = self.parameters.child("Trigger", "Data Time Field").value()
            if data_time_field and data_time_field != "None":
                serializer.set(Scope.TRIGGER_DATA_TIME_FIELD, data_time_field)
        except:
            # Trigger section may not exist in all scope types
            pass


    def _on_save_config_action(self):
        """
        Handle the save configuration action from the File menu.
        Opens a file dialog and saves the current configuration.
        """

        # Open file dialog to get save location
        file_path, _ = QtWidgets.QFileDialog.getSaveFileName(
            self._win,
            "Save Configuration",
            "",
            "C2 Data Viewer Config (*.c2dv);;All Files (*)"
        )

        if file_path:
            try:
                # Ensure file has .c2dv extension
                if not file_path.endswith('.c2dv'):
                    file_path += '.c2dv'

                # Save configuration
                with open(file_path, 'w') as f:
                    self.serialize(f)

                # Show success message
                QtWidgets.QMessageBox.information(
                    self._win,
                    "Success",
                    f"Configuration saved to:\n{file_path}"
                )
            except Exception as e:
                # Show error message
                QtWidgets.QMessageBox.critical(
                    self._win,
                    "Error",
                    f"Failed to save configuration:\n{str(e)}"
                )



# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2020 UChicago Argonne LLC
 as operator of Argonne National Laboratory

Unit tests for Imagev

@author: Matic Pogacnik <mpogacnik@anl.gov>
"""
import os
import sys
import unittest
from pyqtgraph.Qt import QtCore
from pyqtgraph.Qt import QtWidgets

from c2dataviewer.imagev import ImageWindow
from .helper import setup_qt_app

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestImagev(unittest.TestCase):

    def setUp(self):
        """
        Build the environment for each unit test case.
        This method is called before each test.

        :return:
        """

        # Create Qt application
        setup_qt_app()
        
        # Build image window
        self.window = ImageWindow()
        self.window.resize(1400, 1400)
        QtWidgets.QApplication.instance().processEvents()

    def tearDown(self):
        """
        Tear down the environment after each test case.
        This mentod is called after each test.

        :return:
        """

    def test_widgetVisibility2WindowSize(self):
        """
        Test if widgets start to disappear if window
        is resized to smallest possible size.

        As this gets really complex because of the relative coordinates,
        map everything to global (screen) coordinates and test in global space

        """

        # Get main window handle
        mainWindow = self.window

        # Get all child widgets
        widgets = mainWindow.findChildren(QtWidgets.QWidget)

        def pointOnMainWindow(point):

            # Get main window global coordinates
            mainWindowGeometry = mainWindow.geometry()
            mainWindowTopLeft = mainWindow.mapToGlobal(QtCore.QPoint(0, 0))
            mainWindowTopRight = mainWindow.mapToGlobal(QtCore.QPoint(mainWindowGeometry.width(), 0))
            mainWindowBottomLeft = mainWindow.mapToGlobal(QtCore.QPoint(0, mainWindowGeometry.height()))
            #mainWindowBottomRight = mainWindow.mapToGlobal(QPoint(mainWindowGeometry.width(), mainWindowGeometry.height()))

            if (point.x() >= mainWindowTopLeft.x() and
                point.x() <= mainWindowTopRight.x() and
                point.y() >= mainWindowTopLeft.y() and
                point.y() <= mainWindowBottomLeft.y()):
                return True
            else :
                return False

        def testWidgetsIfOnMainScreen():

            # Loop over them and see if they are all contained in a main window
            for _, w in enumerate(widgets):
                # Skip QSizeGrip (Handle which can be clicked with a mouse to resize the window)
                # as this is part of the window itself and is always out of the "canvas"
                if isinstance(w, QtWidgets.QSizeGrip):
                    continue

                self.assertTrue(pointOnMainWindow(w.mapToGlobal(QtCore.QPoint(0, 0))),
                        f"Testing top-left corner of {w.__class__.__name__} widget.")
                self.assertTrue(pointOnMainWindow(w.mapToGlobal(QtCore.QPoint(w.width(),0))),
                        f"Testing top-right corner of {w.__class__.__name__} widget.")
                self.assertTrue(pointOnMainWindow(w.mapToGlobal(QtCore.QPoint(0, w.height()))),
                        f"Testing bottom-left corner of {w.__class__.__name__} widget.")
                self.assertTrue(pointOnMainWindow(w.mapToGlobal(QtCore.QPoint(w.width(), w.height()))),
                        f"Testing bottom-right corner of {w.__class__.__name__} widget.")

        print(f"Testing widget visibility on default window size:")
        print(f"Window width: {mainWindow.geometry().width()}")
        print(f"Window height: {mainWindow.geometry().height()}")
        testWidgetsIfOnMainScreen()



from .config import Parser, Scope

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

class ScopeConfigureBase:
    def __init__(self, params, **kwargs):
        self.params = Parser(params)
        self.default_trigger = kwargs.get("trigger", None)

    def add_source_aquisition_props(self, children, section):
        """
        Add acquisition information specific to the application data source.  This
        should be implemented by the child class
        """
        return children

    def assemble_acquisition(self, section=None, buffer_unit='Samples'):
        """
        Assemble acquisition information

        :param section:
        :return:
        """

        children = []

        buffer_size = self.params.get(Scope.BUFFER)
        
        children += [
            {"name": "Freeze", "type": "bool", "value": False},
            {"name": "Buffer (%s)" % buffer_unit, "type": "int", "value": buffer_size, "siPrefix": False, 'decimals': 20}
        ]

        acquisition = {"name": "Acquisition", "type":"group", "children": children}
        
        return acquisition

    
    def assemble_display(self, section=None, app_section_key=None, default_autoscale=None):
        """
        Assemble display information

        :param section: DISPLAY section in config file
        :param app_section_key: type of app
        :param default_autoscale: default autoscale values for each app
        :return:
        """
        # If AUTOSCALE set in the app specific sections in the config file
        autoscale = self.params.get(Scope.AUTOSCALE, default=default_autoscale)
        display_mode = self.params.get(Scope.DISPLAY_MODE, default='normal')
        fft_filter = self.params.get(Scope.FFT_FILTER, default='none')
        n_average = self.params.get(Scope.AVERAGE, default=1)
        n_average = n_average if n_average > 0 else 1
        single_axis = self.params.get(Scope.SINGLE_AXIS, default=True)
        histogram = self.params.get(Scope.HISTOGRAM, default=False)
        n_binning = self.params.get(Scope.NBIN, default=100)
        refresh = self.params.get(Scope.REFRESH, default=100)

        display = {"name": "Display", "type": "group", "expanded": True,
                   "children": [
                       {"name": "Mode", "type": "list", "limits": {
                           "Normal": "normal",
                           "FFT": "fft",
                           "PSD": "psd",
                           "Diff": "diff",    
                           "Autocorrelation FFT": "autocorrelate_fft",               
                       }, "value": display_mode},
                       {"name": "FFT filter", "type": "list", "limits": {
                           "None" : "none",
                           "Hamming" : "hamming"
                       }, "value": fft_filter},
                       {"name": "Exp moving avg", "type": "int", "value": n_average, "limits" : (1, 1e+10)},
                       {"name": "Autoscale", "type": "bool", "value": autoscale},
                       {"name": "Single axis", "type": "bool", "value": single_axis},
                       {"name": "Histogram", "type": "bool", "value": histogram},
                       {"name": "Num Bins", "type": "int", "value": n_binning},
                       {"name": "Refresh", "type": "float", "value": refresh / 1000., "siPrefix": True, "suffix": "s"},
                   ]}
        
        return display
    

    def assemble_trigger(self, section=None):
        trigger_pv = self.params.get(Scope.TRIGGER, self.default_trigger)
        if trigger_pv is not None and trigger_pv.upper().strip() == "NONE":
            # set trigger PV value to None if a "None" string comes from configuration
            trigger_pv = None

        trigger_mode = self.params.get(Scope.TRIGGER_MODE, 'none')
        if trigger_mode == 'off':
            trigger_mode = 'none'

        config_expanded = True if trigger_mode != "none" else False

        # Load trigger settings from config
        time_field = self.params.get(Scope.TRIGGER_TIME_FIELD, None)
        if time_field is None:
            time_field = "None"

        data_time_field = self.params.get(Scope.TRIGGER_DATA_TIME_FIELD, None)
        if data_time_field is None:
            data_time_field = "None"

        autoscale_buffer = self.params.get(Scope.TRIGGER_AUTOSCALE_BUFFER, True)
        threshold = self.params.get(Scope.TRIGGER_THRESHOLD, 0.0)

        cfg ={"name": "Trigger",
              "type": "group",
              "expanded": config_expanded,
              "children" : [
                  { "name" :  "Mode", "type": "list", "limits": {
                    "Off" : "none",
                    "On change" : "onchange",
                    "Greater than threshold" : "gtthreshold",
                    "Lesser than threshold" : "ltthreshold"
                    },
                  "value" : trigger_mode
                 },
                  {"name": "PV", "type": "str", "value": trigger_pv},
                  {"name": "Trig Status", "type": "str", "value": "", "readonly": True},
                  {"name": "Trig Value", "type": "str", "value": "", "readonly": True},
                  {"name": "Time Field", "type": "list", "limits" : [time_field],
                   "value": time_field, "default" : "None", "visible" : False},
                  {"name": "Data Time Field", "type": "list", "limits" : [data_time_field],
                   "value": data_time_field, "default" : "None" },
                  {"name": "Autoscale Buffer", "type": "bool", "value" : autoscale_buffer},
                  {"name": "Threshold", "type": "float", "value": threshold},
              ]}
        return cfg
    
    def assemble_statistics(self):
        statistics = {"name": "Statistics", "type": "group",
                      "expanded":False,
                      "children": [
            {"name": "CPU", "type": "float", "value": 0, "readonly": True, "suffix": "%"},
            {"name": "Lost Arrays", "type": "int", "value": 0, "readonly": True},
            {"name": "Tot. Arrays", "type": "int", "value": 0, "readonly": True, "siPrefix": True},
            {"name": "Arrays/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Bytes/Sec", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "/sec"},
            {"name": "Rate", "type": "float", "value": 0., "readonly": True, "siPrefix": True,
             "suffix": "Frames/sec"},
        ]}
        return statistics



"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS
"""

import unittest
import os
import sys

from configparser import ConfigParser
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.parametertree import Parameter

from c2dataviewer.model import DataSource as DataReceiver
from c2dataviewer.control.striptoolconfig import StripToolConfigure
from c2dataviewer.striptool import StripToolWindow, WarningDialog, PvEditDialog

from c2dataviewer.control.striptoolconfig import StripToolConfigure
from c2dataviewer.control.striptool_controller import StripToolController
from c2dataviewer.control.pvconfig import PvConfig

os.environ["QT_QPA_PLATFORM"] = "offscreen"

class TestStriptoolController(unittest.TestCase):
    DEFAULT_CFG = """
[STRIPTOOL]
"""

    @classmethod
    def setUpClass(cls):
        """
        Set up QApplication for all tests in this class.
        """
        cls.app = QtWidgets.QApplication.instance()
        if cls.app is None:
            cls.app = QtWidgets.QApplication(sys.argv)

    def setUp(self):
        # Create ImageWindow and get the imageWidget instance
        self.window = StripToolWindow()

        # Build parameter tree
        cf = ConfigParser()
        configure = StripToolConfigure(cf)
        self.parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(self.parameters, showTop=False)

        # Model to be used
        self.model = DataReceiver()

        # Build GUI elements
        self.warning = WarningDialog(None)
        self.pvedit_dialog = PvEditDialog()

    def tearDown(self):
        """
        Tear down the environment after each test case.
        """
        # Clean up the window to prevent Qt object deletion issues
        if hasattr(self, 'window'):
            self.window.close()
            self.window.deleteLater()
        if hasattr(self, 'striptool_controller'):
            del self.striptool_controller
        # Process pending Qt events to ensure cleanup
        QtWidgets.QApplication.processEvents()

    def create_controller(self, configure):
        """
        Helper method to create a controller with the given StripToolConfigure object.

        :param configure: StripToolConfigure object with parsed configuration
        :return: StripToolController instance
        """
        parameters = Parameter.create(
            name="params", type="group", children=configure.parse())
        self.window.parameterPane.setParameters(parameters, showTop=False)

        controller = StripToolController(
            self.window, self.model, self.pvedit_dialog, self.warning, parameters, configure)

        return controller
        

    def test_pv_edit_dialog(self):
        """
        Test PV edit dialog operations.
        """
        raw = """
[STRIPTOOL]
DefaultProtocol = pva
Chan0.PV = Ch0
Chan1.PV = Ch1
Chan2.PV = Ch2
"""
        cfg = ConfigParser()
        cfg.read_string(raw)
        configure = StripToolConfigure(cfg)

        pvcount = 3
        self.striptool_controller = self.create_controller(configure)

        self.assertEqual(len(self.striptool_controller.pvdict), pvcount)
        pvlist = []
        for p in self.striptool_controller.pvdict.values():
            cfg = p.make_pvconfig()
            pvlist.append(cfg)

        pvedit = self.striptool_controller._pvedit_dialog
        pvedit._set_pvlist(pvlist)
        pvedit._add_pv('Ch3', '#000000', 'pva')
        pvedit._on_ok()

        self.assertEqual(len(self.striptool_controller.pvdict), pvcount + 1)
        self.assertTrue('Ch3' in self.striptool_controller.pvdict)

    def test_update_status(self):
        """
        Test update statistics
        """
        cfg = ConfigParser()
        cfg.read_string(self.DEFAULT_CFG)
        configure = StripToolConfigure(cfg)
        self.striptool_controller = self.create_controller(configure)

        #make sure that update_status runs without errors
        self.striptool_controller.update_status()

    # ------------------------------------------------------------------
    # Config file vs keyword arguments tests
    # ------------------------------------------------------------------
    def test_config_display_settings(self):
        """
        Test that display settings from config file are loaded into controller.
        """
        config_str = """
        [STRIPTOOL]
        DefaultProtocol = ca
        AUTOSCALE = False

        [DISPLAY]
        MODE = psd
        AVERAGE = 10
        REFRESH = 300
        HISTOGRAM = True
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)
        configure = StripToolConfigure(cfg)

        controller = self.create_controller(configure)

        # Verify settings are loaded via parameters
        self.assertFalse(controller.parameters.child("Display").child("Autoscale").value())
        self.assertEqual(controller.parameters.child("Display").child("Mode").value(), 'psd')
        self.assertEqual(controller.parameters.child("Display").child("Exp moving avg").value(), 10)
        self.assertEqual(controller.parameters.child("Display").child("Refresh").value(), 0.3)
        self.assertTrue(controller.parameters.child("Display").child("Histogram").value())

    def test_config_pvs_loaded(self):
        """
        Test that PVs from config are properly loaded into controller's pvdict.
        """
        config_str = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = TEST:PV:ONE
        Chan1.Color = #FF0000
        Chan2.PV = pva://TEST:PV:TWO
        Chan2.Color = #00FF00
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)
        configure = StripToolConfigure(cfg)

        controller = self.create_controller(configure)

        # Verify PVs are loaded into controller
        self.assertEqual(len(controller.pvdict), 2)
        self.assertIn('TEST:PV:ONE', controller.pvdict)
        self.assertIn('TEST:PV:TWO', controller.pvdict)

        # Verify PV details
        pv1_config = controller.pvdict['TEST:PV:ONE'].make_pvconfig()
        self.assertEqual(pv1_config.color, '#FF0000')

        pv2_config = controller.pvdict['TEST:PV:TWO'].make_pvconfig()
        self.assertEqual(pv2_config.color, '#00FF00')

    def test_default_protocol(self):
        """
        Test that default protocol from config is used by controller.
        """
        from c2dataviewer.model import make_protocol

        config_str = """
        [STRIPTOOL]
        DefaultProtocol = pva
        Chan1.PV = TEST:PV:DEFAULT
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)
        configure = StripToolConfigure(cfg)

        controller = self.create_controller(configure)

        # Verify default protocol is used
        self.assertEqual(configure.default_proto, 'pva')
        pv_config = controller.pvdict['TEST:PV:DEFAULT'].make_pvconfig()
        self.assertEqual(pv_config.proto, make_protocol('pva'))

    def test_pv_kwarg_integration(self):
        """
        Test that PVs passed via kwargs are integrated into controller.
        """
        cfg = ConfigParser()
        cfg.read_string("[STRIPTOOL]\nDefaultProtocol = ca")

        # Pass PV via kwarg
        configure = StripToolConfigure(cfg, pv={'kwarg_pv': 'KWARG:TEST:PV'})

        controller = self.create_controller(configure)

        # Verify kwarg PV is in controller
        self.assertIn('KWARG:TEST:PV', controller.pvdict)

    def test_mixed_config_and_kwarg_pvs(self):
        """
        Test controller with both config file PVs and kwarg PVs.
        """
        config_str = """
        [STRIPTOOL]
        DefaultProtocol = ca
        Chan1.PV = CONFIG:PV
        Chan1.Color = #0000FF
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)

        # Add kwarg PV
        configure = StripToolConfigure(cfg, pv={'kwarg': 'KWARG:PV'})

        controller = self.create_controller(configure)

        # Both PVs should be in controller
        self.assertIn('CONFIG:PV', controller.pvdict)
        self.assertIn('KWARG:PV', controller.pvdict)
        self.assertEqual(len(controller.pvdict), 2)

        # Verify config PV retains its color
        config_pv = controller.pvdict['CONFIG:PV'].make_pvconfig()
        self.assertEqual(config_pv.color, '#0000FF')
        
    def test_serialize_full_config(self):
        """
        Test that serialize function writes proper configuration to file with PVs.
        """
        from io import StringIO

        config_str = """
        [STRIPTOOL]
        DefaultProtocol = pva
        Chan1.PV = TEST:PV:ONE
        Chan1.Color = #FF0000
        Chan2.PV = ca://TEST:PV:TWO
        Chan2.Color = #00FF00
        Chan3.PV = TEST:PV:THREE
        Chan3.Color = #0000FF
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)
        configure = StripToolConfigure(cfg)

        controller = self.create_controller(configure)

        # Serialize to StringIO
        output = StringIO()
        controller.serialize(output)

        # Read back the serialized config
        output.seek(0)
        result_cfg = ConfigParser()
        result_cfg.read_file(output)

        # Verify app type
        self.assertEqual(result_cfg.get('DEFAULT', 'APP'), 'STRIPTOOL')

        # Verify PVs are serialized with protocol prefixes
        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan1.pv'), 'pva://TEST:PV:ONE')
        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan1.color'), '#FF0000')

        # Chan2 already has ca:// prefix in config, should keep it
        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan2.pv'), 'ca://TEST:PV:TWO')
        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan2.color'), '#00FF00')

        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan3.pv'), 'pva://TEST:PV:THREE')
        self.assertEqual(result_cfg.get('STRIPTOOL', 'chan3.color'), '#0000FF')

        # Verify scope config is serialized
        self.assertTrue(result_cfg.has_section('DISPLAY'))
        self.assertTrue(result_cfg.has_section('ACQUISITION'))

        # Verify sample mode is serialized
        self.assertTrue(result_cfg.has_option('ACQUISITION', 'SAMPLEMODE'))

    def test_serialize_minimal_config(self):
        """
        Test serialize with minimal configuration (no PVs).
        """
        from io import StringIO

        config_str = """
        [STRIPTOOL]
        DefaultProtocol = ca
        """
        cfg = ConfigParser()
        cfg.read_string(config_str)
        configure = StripToolConfigure(cfg)

        controller = self.create_controller(configure)

        # Serialize to StringIO
        output = StringIO()
        controller.serialize(output)

        # Read back the serialized config
        output.seek(0)
        result_cfg = ConfigParser()
        result_cfg.read_file(output)

        # Verify app type
        self.assertEqual(result_cfg.get('DEFAULT', 'APP'), 'STRIPTOOL')

        # Verify STRIPTOOL section exists even with no PVs
        self.assertTrue(result_cfg.has_section('STRIPTOOL'))


if __name__ == '__main__':
    unittest.main()





# -*- coding: utf-8 -*-

"""
C2 DATA VIEWER is distributed subject to a Software License Agreement found
in the file LICENSE that is included with this distribution.
SPDX-License-Identifier: EPICS

Copyright 2018 UChicago Argonne LLC
 as operator of Argonne National Laboratory

PVA object viewer utilities

@author: Guobao Shen <gshen@anl.gov>
"""

import sys
import os.path
import psutil
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import uic
from pyqtgraph import QtCore
from .control import ImageController
from .model import DataSource as DataReceiver


form_path = os.path.join(os.path.dirname(__file__), "ui/imagev.ui")
form_class = uic.loadUiType(form_path)[0]

class ImageWindow(QtWidgets.QMainWindow, form_class):
    resized = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        """

        :param parent:
        """
        super(ImageWindow, self).__init__(parent=parent)
        self._proc = psutil.Process()
        self.setupUi(self)
        self.show()

        self.resized.connect(self.resizedCallback)

    def resizeEvent(self, event):
        """

        :param event:
        :return:
        """
        self.resized.emit()
        return super(ImageWindow, self).resizeEvent(event)

    def resizedCallback(self):
        """

        :return:
        """
        self.imageWidget.set_scaling()


settings_dialog_path = os.path.join(os.path.dirname(__file__), "ui/imagev_settings.ui")
settings_dialog_class = uic.loadUiType(settings_dialog_path)[0]

class ImageSettingsDialog(QtWidgets.QDialog, settings_dialog_class):
    def __init__(self, parent=None):
        super(ImageSettingsDialog, self).__init__(parent=parent)
        self.setupUi(self)


warning_path = os.path.join(os.path.dirname(__file__), "ui/warning.ui")
warning_class = uic.loadUiType(warning_path)[0]

class WarningDialog(QtWidgets.QDialog, warning_class):
    def __init__(self, parent=None):
        super(WarningDialog, self).__init__(parent=parent)
        self.setupUi(self)

def imagev(pv, scale=None, noAGC=True):
    """
    Main function for image display

    :param pv: dictionary, which host all pv names
    :param label: label to identify PVs
    :param scale: scaling factor relative to 800xYYYY, 1.0 by default
    :param noAGC: No automatic gain adjust, True by default
    :return:
    """

    # Check for an instance of a QtWidgets.QApplication, if so use it...
    app = QtWidgets.QApplication.instance()
    if app is None:
        app = QtWidgets.QApplication(sys.argv)
    else:
        print('QApplication instance already exists: %s' % str(app))

    if pv is None:
        label = None
    else:
        label = list(pv.keys())
        
    w = ImageWindow(None)
    data = DataReceiver(QtCore.QTimer, default=None)
    w.imageWidget.set_datasource(data)

    settings_dialog = ImageSettingsDialog(None)
    warning = WarningDialog(None)
    ImageController(w, IMAGE_SETTINGS_DIALOG = settings_dialog, WARNING=warning,
                    PV=label, timer=QtCore.QTimer(), data=data)

    # set initial scaling factor
    if scale is not None and scale != 1.0:
        w.imageWidget.set_scaling(scale=scale)
    else:
        w.imageWidget.set_scaling()

    if not noAGC:
        w.imageWidget.enable_auto_gain()

    if (label == None):
        msg = "No input channel specified, please add it manually in the Camera textbox."
        warning.warningTextBrowse.setText(msg)
        warning.show()
        warning.warningConfirmButton.clicked.connect(warning.close)

    w.show()
    sys.exit(app.exec_())
